//R: here kernels for SETI MultiBeam OpenCL build will be located
// This version is using fraction(a * b * b) = fraction((_a + fa) (_b + fb) ^ 2) for enhancing the precision
#define M_2PI              6.28318530717958647692f
__kernel __attribute__((vec_type_hint(float4)))
void CalcChirpData_kernel2_cl(const float4 chirp_rate, __global float4* cx_DataArray,
									   __global float4* cx_ChirpDataArray,__global float4* sample_rate) {  
	uint i= get_global_id(0);   
	float4 cx = cx_DataArray[i];
	float4 time = sample_rate[i];
	float sin1,sin2, cos1,cos2;
	float4 chirped;
	float tmp;
	float a = chirp_rate.x;
	float b = time.x;
	float a1= chirp_rate.z;
	float b1= time.z;
	float fa = chirp_rate.y;
	float fb = time.y;     
	float fa1 = chirp_rate.w;
	float fb1 = time.w;     
	float ang  = fract(2 * a * fract(b * fb,&tmp),&tmp) + fract(a * fb * fb,&tmp) + 
		fract(b * fract(b * fa,&tmp),&tmp) + fract(2 * b * fa * fb,&tmp) + fa * fb * fb;
	ang -= floor(ang);
	ang *= M_2PI;
	//sin1=sincos(ang,&cos1); 
	sin1=native_sin(ang);
	cos1=native_cos(ang);
	chirped.x = cx.x * cos1 - cx.y * sin1;
	chirped.y = cx.x * sin1 + cx.y * cos1;
	ang  = fract(2 * a1 * fract(b1 * fb1,&tmp),&tmp) + fract(a1 * fb1 * fb1,&tmp) + 
		fract(b1 * fract(b1 * fa1,&tmp),&tmp) + fract(2 * b1 * fa1 * fb1,&tmp) + fa1 * fb1 * fb1;
	ang -= floor(ang);
	ang *= M_2PI;
	//sin2=sincos(ang,&cos2); 
	sin2=native_sin(ang);
	cos2=native_cos(ang);
	chirped.z = cx.z * cos2 - cx.w * sin2;
	chirped.w = cx.z * sin2 + cx.w * cos2;

		cx_ChirpDataArray[i] = chirped;

}
#if 0 //ATI kernel
__kernel __attribute__((vec_type_hint(float4)))
__kernel void GetPowerSpectrum_kernel_cl(__global float4* FreqData, __global float4* PowerSpectrum) {
//R: will do 4 dots at once for better memory access on ATI GPUs
	uint i= get_global_id(0);   
	float4 f1 = FreqData[2*i];
	float4 f2 = FreqData[2*i+1];
	float4 p;
	p.xy=mad(f1.xz,f1.xz,f1.yw*f1.yw);
	p.zw=mad(f2.xz,f2.xz,f2.yw*f2.yw);
	PowerSpectrum[i] = p;
}
#elif 1
__kernel void GetPowerSpectrum_kernel_cl(__global float4* FreqData, __global float2* PowerSpectrum) {
	uint i= get_global_id(0);   
	float4 f1 = FreqData[i];
	float2 p;
	p.xy=mad(f1.xz,f1.xz,f1.yw*f1.yw);
	PowerSpectrum[i] = p;
}
#else //NV kernel
__kernel void GetPowerSpectrum_kernel_cl(__global float2* FreqData, __global float* PowerSpectrum) {
	uint i= get_global_id(0);   
	float2 f1 = FreqData[i];
	float p;
	p=mad(f1.x,f1.x,f1.y*f1.y);
	PowerSpectrum[i] = p;
}
#endif

__kernel __attribute__((vec_type_hint(float4)))
void PC_find_spike_kernel_cl(__global float4* PowerSpectrum, __global float4* MeanMaxIdx, uint fftlen,
									  float best, float spike_thresh,__global uint4* result_flag){
	//R: CUDA version of this kernel use shared memory and reducing, HD4xxx have no shared memory so I will
	//do whole power array per workitem.
	uint tid= get_global_id(0);
	float mean=0.f;
	float4 sum4=0.f;
	float maxval=0.f;
	float pos=0.f;//R: should be integer value but will use float for better memory packing
	float4 mmi;
	float4 p;
	//R: separate handling of first quad - DC element should be omitted
	p=PowerSpectrum[tid*(fftlen>>2)];
	sum4.x=p.x;
	sum4.y=p.y;
	sum4.z=p.z;
	sum4.w=p.w;
	if(p.y>maxval){pos=1;}
	maxval=max(maxval,p.y);
	if(p.z>maxval){pos=2;}
	maxval=max(maxval,p.z);
	if(p.w>maxval){pos=3;}
	maxval=max(maxval,p.w);
	for(uint i=1;i<(fftlen>>2);i++){
		p=PowerSpectrum[i+tid*(fftlen>>2)];//R:doing 4 elements at once
		sum4+=p;
		if(p.x>maxval){pos=i*4;}
		maxval=max(maxval,p.x);
		if(p.y>maxval){pos=i*4+1;}
		maxval=max(maxval,p.y);
		if(p.z>maxval){pos=i*4+2;}
		maxval=max(maxval,p.z);
		if(p.w>maxval){pos=i*4+3;}
		maxval=max(maxval,p.w);
	}
	mean=(sum4.x+sum4.y+sum4.z+sum4.w)/fftlen;
	mmi.x=mean;
	mmi.y=maxval;
	mmi.z=pos;
	//R: w field not used
	MeanMaxIdx[tid]=mmi;
	float power=maxval/mean;
	float score=log10(power/40.f);
	if(score>best || power>spike_thresh){
		result_flag[0].x=1;//R: only in this case kernel results are needed
	}
}
__kernel void PC_find_spike8_kernel_cl(__global float* PowerSpectrum, __global float4* MeanMaxIdx,
										float best, float spike_thresh,__global uint4* result_flag,__local float4* tmp){
//R:each block working with few arrays (8 for spike8) at once (but uses only first dim for this)
	const int tid = get_local_id(0);
    const int y = get_global_id(1);
    //const int gridX = get_local_size(0);
    const int start = y * 64;//gridX;
    
    //=8-1=7 const int n1 = n - 1;
    //=64/8=8 const int width = blockDim.x / n;
    
    float val = PowerSpectrum[start + tid];
	if ((tid & 7) == 0)
		tmp[tid] =(float4)(val, 0.0f, (float)(tid&7),0.f); // Original max omits first element of every fft
	else
		tmp[tid] = (float4)(val, val, (float)(tid&7),0.f);

    for (int i = 4; i > 0; i >>= 1) {
        barrier(CLK_LOCAL_MEM_FENCE); 
        if ((tid & 7) < i) {
            tmp[tid] = (float4)(
                tmp[tid].x + tmp[tid + i].x, 
                max(tmp[tid].y, tmp[tid + i].y),
                tmp[tid].y > tmp[tid + i].y ? tmp[tid].z : tmp[tid + i].z,0.f); 
        }
    }
    barrier(CLK_LOCAL_MEM_FENCE); 


    if (tid < 8) {
		float4 tmp4;
		tmp4=tmp[tid<<3];
		tmp4.x/=8.f;//mean value needed, not just sum; sum/fftlen
        MeanMaxIdx[(y<<3)+ tid] = tmp4;
		float power=tmp4.y/tmp4.x;//max/mean
		float score=log10(power/40.f);
		if(score>best || power>spike_thresh){
			result_flag[0].x=1;//R: only in this case kernel results are needed
		}
    }
}
__kernel void PC_find_spike16_kernel_cl(__global float* PowerSpectrum, __global float4* MeanMaxIdx,
										float best, float spike_thresh,__global uint4* result_flag,__local float4* tmp){
//R:each block working with few arrays (4 for spike16) at once (but uses only first dim for this)
	const int tid = get_local_id(0);
    const int y = get_global_id(1);
    //const int gridX = get_local_size(0);
    const int start = y * 64;//gridX;
    
    //=16-1=15 const int n1 = n - 1;
    //=64/16=4 const int width = blockDim.x / n;
    
    float val = PowerSpectrum[start + tid];
	if ((tid & 15) == 0)
		tmp[tid] =(float4)(val, 0.0f, (float)(tid & 15),0.f); // Original max omits first element of every fft
	else
		tmp[tid] = (float4)(val, val, (float)(tid & 15),0.f);

    for (int i = 8; i > 0; i >>= 1) {
        barrier(CLK_LOCAL_MEM_FENCE); 
        if ((tid & 15) < i) {
            tmp[tid] = (float4)(
                tmp[tid].x + tmp[tid + i].x, 
                max(tmp[tid].y, tmp[tid + i].y),
                tmp[tid].y > tmp[tid + i].y ? tmp[tid].z : tmp[tid + i].z,0.f); 
        }
    }
    barrier(CLK_LOCAL_MEM_FENCE); 


    if (tid < 4) {
		float4 tmp4;
		tmp4=tmp[tid<<4];
		tmp4.x/=16.f;//mean value needed, not just sum; sum/fftlen
        MeanMaxIdx[(y<<2)+ tid] = tmp4;
		float power=tmp4.y/tmp4.x;//max/mean
		float score=log10(power/40.f);
		if(score>best || power>spike_thresh){
			result_flag[0].x=1;//R: only in this case kernel results are needed
		}
    }
}
#if USE_OPENCL_NV || USE_OPENCL_HD5xxx
__kernel void PC_find_spike32_kernel_cl(__global float* PowerSpectrum, __global float4* MeanMaxIdx, uint Nchunk,
									  float best, float spike_thresh,__global uint4* result_flag,__local float4* tmp){
//R: float4 in shared mem is waste of resources, but CUDA SDK3.2RC2 doesn't understand float3... weird...    
	const int tid = get_local_id(0);
    const int y = get_global_id(1);
    const int gridX = get_local_size(0);
    const int start = y * gridX * Nchunk;
    const int end = (y + 1) * gridX*Nchunk;
	
    float sum = 0.0f;
    float maximum = 0.0f;    
    int pos = 0;

	int i = start + tid;
	float val = PowerSpectrum[i]; 
	if (tid > 0) { // Original max omits first element of every fft
		sum += val;
		if (val > maximum)
			pos = i;
		maximum = max(maximum, val);
	}

    for (i = start + tid + gridX; i < end; i += gridX) {
        float val = PowerSpectrum[i];
        sum += val;
        if (val > maximum)
            pos = i;
        maximum = max(maximum, val);
    }
	
    tmp[tid] =(float4)(sum, maximum, (float)(pos - start),0.f);

    for (i = gridX >> 1; i > 0; i >>= 1) {
        barrier(CLK_LOCAL_MEM_FENCE); 
        if (tid < i) {            
            tmp[tid] = (float4)(
                tmp[tid].x + tmp[tid + i].x, 
                max(tmp[tid].y, tmp[tid + i].y),
                tmp[tid].y > tmp[tid + i].y ? tmp[tid].z : tmp[tid + i].z,0.f//R:w not used
                ); 
        }
    }
    barrier(CLK_LOCAL_MEM_FENCE); 
    if (tid == 0) {
		float4 tmp4;
		tmp4.xyzw=tmp[tid].xyzw;//R:float4 in output array but float3 in shared memory
		tmp4.x/=(gridX * Nchunk);//mean value needed, not just sum; sum/fftlen
        MeanMaxIdx[y] = tmp4;
		float power=tmp4.y/tmp4.x;//max/mean
		float score=log10(power/40.f);
		if(score>best || power>spike_thresh){
			result_flag[0].x=1;//R: only in this case kernel results are needed
		}
	}

}
#endif
__kernel __attribute__((vec_type_hint(float4)))
void PC_find_spike_reduce0_kernel_cl(__global float4* PowerSpectrum, __global float4* tmp, uint fftlen){
	//R: CUDA version of this kernel use shared memory and reducing, HD4xxx have no shared memory so I will
	//do whole power array per workitem.
	uint tid= get_global_id(0);
	uint y=get_global_id(1);
	float4 sum4=(float4)0.f;
	float maxval=0.f;
	float pos=0.f;//R: should be integer value but will use float for better memory packing
	float4 sum_val_pos;
	float4 p;
	int i=0;

	//R: separate handling of first quad - DC element should be omitted
	if(y==0){
		p=PowerSpectrum[tid*(fftlen>>2)];
		sum4.x=p.x;
		sum4.y=p.y;
		sum4.z=p.z;
		sum4.w=p.w;
		if(p.y>maxval){pos=1;}
		maxval=max(maxval,p.y);
		if(p.z>maxval){pos=2;}
		maxval=max(maxval,p.z);
		if(p.w>maxval){pos=3;}
		maxval=max(maxval,p.w);
		i=1;
	}
	for(;i<(64);i++){
		p=PowerSpectrum[i+y*64+tid*(fftlen>>2)];//R:doing 4 elements at once
		sum4+=p;
		if(p.x>maxval){pos=i*4;}
		maxval=max(maxval,p.x);
		if(p.y>maxval){pos=i*4+1;}
		maxval=max(maxval,p.y);
		if(p.z>maxval){pos=i*4+2;}
		maxval=max(maxval,p.z);
		if(p.w>maxval){pos=i*4+3;}
		maxval=max(maxval,p.w);
	}
	sum_val_pos.x=sum4.x+sum4.y+sum4.z+sum4.w;
	sum_val_pos.y=maxval;
	sum_val_pos.z=pos+256*y;
	//R: w field not used
	tmp[tid*(fftlen>>8)+y]=sum_val_pos;
}
__kernel __attribute__((vec_type_hint(float4)))
void PC_find_spike_reduce1_kernel_cl(__global float4* tmp, __global float4* MeanMaxIdx, uint fftlen){
	//R: CUDA version of this kernel use shared memory and reducing, HD4xxx have no shared memory so I will
	//do whole power array per workitem.
	uint tid= get_global_id(0);
	float mean=0.f;
	float sum=0.f;
	float maxval=0.f;
	float pos=0.f;//R: should be integer value but will use float for better memory packing
	float4 mmi;
	float4 sum_val_pos;
	for(int y=0;y<(fftlen>>8);y++){
		sum_val_pos=tmp[y+tid*(fftlen>>8)];
		sum+=sum_val_pos.x;
		if(sum_val_pos.y>maxval){pos=sum_val_pos.z;}
		maxval=max(maxval,sum_val_pos.y);
	}
	mean=sum/fftlen;
	mmi.x=mean;
	mmi.y=maxval;
	mmi.z=pos;
	//R: w field not used
	MeanMaxIdx[tid]=mmi;
	//float power=maxval/mean;
	//float score=log10(power/40);
	//if(score>best || power>spike_thresh){
	//	result_flag[0].x=1;//R: only in this case kernel results are needed
	//}
}

#if 0
__kernel void ComputeMeanMaxFromFreq_kernel_cl(__global float4* FreqData, __global float4* MeanMaxIdx, uint fftlen){
	//R: CUDA version of this kernel use shared memory and reducing, HD4xxx have no shared memory so I will
	//do whole power array per workitem.
	uint tid= get_global_id(0);
	float mean=0.f;
	float4 sum4=0.f;
	float maxval=0.f;
	float pos=0.f;//R: should be integer value butwill use float for better memory packing
	float4 mmi;
	//R: separate handling of first quad - DC element should be omitted
	float4 f1 = FreqData[tid*(fftlen>>1)];
	float4 f2 = FreqData[tid*(fftlen>>1)+1];
	float4 p;
	p.xy=f1.xz*f1.xz+f1.yw*f1.yw;
	p.zw=f2.xz*f2.xz+f2.yw*f2.yw;
	sum4.x=0.f;
	sum4.y=p.y;
	sum4.z=p.z;
	sum4.w=p.w;
	if(p.y>maxval){pos=1;}
	maxval=max(maxval,p.y);
	if(p.z>maxval){pos=2;}
	maxval=max(maxval,p.z);
	if(p.w>maxval){pos=3;}
	maxval=max(maxval,p.w);
	for(uint i=1;i<(fftlen>>2);i++){
		f1 = FreqData[tid*(fftlen>>1)+2*i];
		f2 = FreqData[tid*(fftlen>>1)+1+2*i];
		p.xy=f1.xz*f1.xz+f1.yw*f1.yw;
		p.zw=f2.xz*f2.xz+f2.yw*f2.yw;
		sum4+=p;
		if(p.x>maxval){pos=i*4;}
		maxval=max(maxval,p.x);
		if(p.y>maxval){pos=i*4+1;}
		maxval=max(maxval,p.y);
		if(p.z>maxval){pos=i*4+2;}
		maxval=max(maxval,p.z);
		if(p.w>maxval){pos=i*4+3;}
		maxval=max(maxval,p.w);
	}
	mean=(sum4.x+sum4.y+sum4.z+sum4.w)/fftlen;
	mmi.x=mean;
	mmi.y=maxval;
	mmi.z=pos;
	//R: w field not used
	MeanMaxIdx[tid]=mmi;
}
#endif
__kernel __attribute__((vec_type_hint(float4)))
void Transpose4_kernel_cl(__global float4* in, __global float4* out,uint x_size,uint y_size){
	//R: ATI HD4xxx has no cache exposed in OpenCL so do single possible memory optimization - vector reads in register file.
	//R: x_size matrix width, y_size - matrix heigh
	uint tidx= get_global_id(0);
	uint tidy= get_global_id(1);
	float4 d0=in[tidx+(x_size>>2)*(tidy<<2)];
	float4 d1=in[tidx+(x_size>>2)*((tidy<<2)+1)];
	float4 d2=in[tidx+(x_size>>2)*((tidy<<2)+2)];
	float4 d3=in[tidx+(x_size>>2)*((tidy<<2)+3)];
	float4 o0,o1,o2,o3;
	o0.x=d0.x;o0.y=d1.x;o0.z=d2.x;o0.w=d3.x;
	o1.x=d0.y;o1.y=d1.y;o1.z=d2.y;o1.w=d3.y;
	o2.x=d0.z;o2.y=d1.z;o2.z=d2.z;o2.w=d3.z;
	o3.x=d0.w;o3.y=d1.w;o3.z=d2.w;o3.w=d3.w;
	out[tidy+(y_size>>2)*(tidx<<2)]=o0;
	out[tidy+(y_size>>2)*((tidx<<2)+1)]=o1;
	out[tidy+(y_size>>2)*((tidx<<2)+2)]=o2;
	out[tidy+(y_size>>2)*((tidx<<2)+3)]=o3;
}
__kernel __attribute__((vec_type_hint(float4)))
void GetFixedPoT_kernel_cl(__global float4* power, __global float4* PoT,int fftlen) {
	int ul_PoT_i  = get_global_id(0);//R: dots position in single PoT array (0 to 64)
	int ul_PoT = get_global_id(1);//R: index of PoT array (there are fflen/4 (each workitem handles 1 dot for 4 PoT arrays) number of such arrays, each 64 points wide)

	int   
		ul_PoTChunkSize,
		ul_PoTChunk_i,
		ul_PoTChunkLimit,
		tPowerLen;//R: length of transposed power array

	tPowerLen = 1024*1024 / fftlen;//R: we have 1M of dots in initial power 2D matrix
	ul_PoTChunkSize  = tPowerLen / 64;//R: for now PoT array size for Gauss fitting is 64, this size in dots, not dots4

	// If the number of spectra is greater than the number
	// of elements in a PoT array, we add sum adjacent spectra
	// into PoT elements.
	// ul_PoTChunkSize indicates how many time-wise
	// power spectra bins are added together to make one PoT bin.

	ul_PoTChunk_i = ul_PoTChunkSize * ul_PoT_i;
	ul_PoTChunkLimit = ul_PoTChunk_i + ul_PoTChunkSize;

	float4 sum = (float4)0.0f;
	for (; ul_PoTChunk_i < ul_PoTChunkLimit; ul_PoTChunk_i++) {
		sum += power[(fftlen>>2)*ul_PoTChunk_i+ul_PoT];
	}
	PoT[ul_PoT_i*(fftlen>>2)+ul_PoT] = sum;
}
//R: TODO check if pre-check should be added in Normalize kernel to skip some areas of PoT matrix
__kernel __attribute__((vec_type_hint(float4)))
void NormalizePoT_kernel_cl(int ul_FftLength,__global float4* PoT,
									 __global float4* PoTPrefixSum, __global float4* NormMaxPower) {
	int tid= get_global_id(0);//R: should be from 0 to ul_FftLength/4 
//R: do 4 PoT arrays at once
//TODO: try to allocate private array of 64 float4 elements in register file to eliminate secondary read from
//global memory. But check of assembly output required to make sure private array not spilled into global memory
	float4 TotalPower = (float4)0.0f;
	for (int i = 0; i < 64; i++) {
		TotalPower += PoT[i * (ul_FftLength>>2)+tid];
	}
	float4 MeanPower = TotalPower / (float4)64;

	// Normalize power-of-time
	float4 sum = (float4)0.0f;
	float4 NormMaxPower_i = (float4)0.0f;
	for (int i = 0; i < 64; i++) {
		float4 PoTi = PoT[i * (ul_FftLength>>2)+tid] /MeanPower;
		NormMaxPower_i = max(NormMaxPower_i, PoTi);
		sum += PoTi;
		PoT[i * (ul_FftLength>>2)+tid] = PoTi;
		PoTPrefixSum[i * (ul_FftLength>>2)+tid] = sum;
	}
	NormMaxPower[tid] = NormMaxPower_i;
}



typedef struct {  
    int iSigma;
//    int gauss_pot_length;
    float GaussPowerThresh;
    float GaussPeakPowerThresh3;
    float GaussSigmaSq;
    int GaussTOffsetStart;
    int GaussTOffsetStop;
    float PeakScaleFactor;
    float GaussChiSqThresh;
    float gauss_null_chi_sq_thresh;
	float gauss_chi_sq_thresh;
//	float* dev_PoT;
//    float* dev_PoTPrefixSum;
//    float* dev_PowerSpectrum;
//    float4* dev_GaussFitResults;
//    float4* dev_GaussFitResultsReordered;
//    float4* dev_GaussFitResultsReordered2;
//    float* dev_NormMaxPower;    
//    float* dev_outputposition;
	float score_offset;
//    int NumDataPoints;
//	float f_weight[CUDA_ACC_MAX_GaussTOffsetStop]; // cached  static_cast<float>(EXP(i, 0, PoTInfo.GaussSigmaSq));
//    unsigned int nsamples;
} ocl_GaussFit_t;
__kernel __attribute__((vec_type_hint(float4)))
void NormalizePoT_peak_PC_kernel_cl(int ul_FftLength,__global float4* PoT,
									 __global float4* PoTPrefixSum, __global float4* NormMaxPower,
									 __constant ocl_GaussFit_t* settings,__global uint* WeakPeaks) {
	int tid= get_global_id(0);//R: should be from 0 to ul_FftLength/4 
//R: do 4 PoT arrays at once
//TODO: try to allocate private array of 64 float4 elements in register file to eliminate secondary read from
//global memory. But check of assembly output required to make sure private array not spilled into global memory
	float4 TotalPower = (float4)0.0f;
	for (int i = 0; i < 64; i++) {
		TotalPower += PoT[i * (ul_FftLength>>2)+tid];
	}
	float4 MeanPower = TotalPower / (float4)64;

	// Normalize power-of-time
	float4 sum = (float4)0.0f;
	float4 NormMaxPower_i = (float4)0.0f;
	for (int i = 0; i < 64; i++) {
		float4 PoTi = PoT[i * (ul_FftLength>>2)+tid] /MeanPower;
		NormMaxPower_i = max(NormMaxPower_i, PoTi);
		sum += PoTi;
		PoT[i * (ul_FftLength>>2)+tid] = PoTi;
		PoTPrefixSum[i * (ul_FftLength>>2)+tid] = sum;
	}
	NormMaxPower[tid] = NormMaxPower_i;
	float thresh=settings->GaussPowerThresh;
	if(NormMaxPower_i.x>thresh ||NormMaxPower_i.y>thresh ||
		NormMaxPower_i.z>thresh || NormMaxPower_i.w>thresh)	WeakPeaks[tid]=1;
	else WeakPeaks[tid]=0;
}


float gammln(float a) {
	float x,y,tmp,ser;
	float cof[6]={76.18009172947146f,-86.50532032941677f,
		24.01409824083091f,-1.231739572450155f,
		0.1208650973866179e-2f,-0.5395239384953e-5f};

	y=x=a;
	tmp=x+5.5f;
	tmp -= (x+0.5f)*log(tmp);
	ser=1.000000000190015f;
	for (int j=0;j<=5;j++){y+=1.f; ser += cof[j]/y;}
	return (float)(-tmp+log(2.5066282746310005f*ser/x));
}
#define ITMAX 10000  // Needs to be a few times the sqrt of the max. input to lcgf
//R: this function called only from device
//R: wlll process 4 PoTs at once as whole ATI OpenCL gauss search chain does - memory access optimization and better computing density
float4 GetTrueMean2(__global float4* fp_PoTPrefixSum, int ul_TOffset, int ul_ExcludeLen, int ul_FftLength,int tid) {
	// TrueMean is the mean power of the data set minus all power
	// out to ExcludeLen from our current TOffset.
	int i_start, i_lim;
	float4 f_ExcludePower;

	// take care that we do not add to exclude power beyond PoT bounds!
	i_start = max(ul_TOffset - ul_ExcludeLen, 0) - 1;
	i_lim = min(ul_TOffset + ul_ExcludeLen + 1, 64) - 1;

	f_ExcludePower = fp_PoTPrefixSum[i_lim * (ul_FftLength>>2)+tid];
	if (i_start >= 0)
		f_ExcludePower -= fp_PoTPrefixSum[i_start * (ul_FftLength>>2)+tid]; 

	return(((float4)(64.0f) - f_ExcludePower) / (float4)(64 - (i_lim - i_start)));
}
float4 GetPeak(__global float4* fp_PoT, int ul_TOffset, int ul_HalfSumLength, 
					   float4 f_MeanPower, float f_PeakScaleFactor, int ul_FftLength,int tid,__constant float* f_weight) {
	// Peak power is calculated as the weighted
	// sum of all powers within ul_HalfSumLength
	// of the assumed gaussian peak.
	// The weights are given by the gaussian function itself.
	// BUG WATCH : for the f_PeakScaleFactor to work,
	// ul_HalfSumLength *must* be set to sigma.

	int i;
	float4 f_sum=(float4)0.0f;

	// Find a weighted sum
	for (i = ul_TOffset - ul_HalfSumLength; i <= ul_TOffset + ul_HalfSumLength; i++) {
		f_sum += (fp_PoT[i * (ul_FftLength>>2)+tid] - f_MeanPower) * f_weight[abs(i-ul_TOffset)];
	}
	return(f_sum * f_PeakScaleFactor);
}

float4 GetChiSq(__global float4* fp_PoT, int ul_FftLength,int ul_TOffset, 
				float4 f_PeakPower, float4 f_MeanPower, float4* xsq_null,int tid,__constant float* f_weight) {
	// We calculate our assumed gaussian powers
	// on the fly as we try to fit them to the
	// actual powers at each point along the PoT.

	float4 f_ChiSq = (float4)0.0f, f_null_hyp=(float4)0.0f;
	float rebin = 1024*1024 / ul_FftLength / 64;

	float4 recip_MeanPower = (float4)1.0f / f_MeanPower;	
	for (int i = 0; i < 64; ++i) {				
		float4 f_PredictedPower = f_MeanPower + f_PeakPower * (float4)f_weight[abs(i - ul_TOffset)];		
		f_PredictedPower *= recip_MeanPower;

		// ChiSq in this realm is:
		//  sum[0:i]( (observed power - expected power)^2 / expected variance )
		// The power of a signal is:
		//  power = (noise + signal)^2 = noise^2 + signal^2 + 2*noise*signal
		// With mean power normalization, noise becomes 1, leaving:
		//  power = signal^2 +or- 2*signal + 1
		float4 noise=2.0f*sqrt(max(f_PredictedPower,1.0f)-1.0f)+1.0f;
		float4 recip_noise = (float4)rebin / noise;

		float4 PoTval = fp_PoT[i * (ul_FftLength>>2)+tid] * recip_MeanPower;
		f_ChiSq += (recip_noise*(PoTval - f_PredictedPower)*(PoTval - f_PredictedPower));
		f_null_hyp+= (recip_noise*(PoTval - 1.0f)*(PoTval - 1.0f));
	}

	f_ChiSq/=(float4)(64.0f);
	f_null_hyp/=(float4)(64.0f);

	(*xsq_null)=f_null_hyp;
	return f_ChiSq;
}
float4 lcgf(float a, float4 x) {
	const float EPS= 1.19209e-006f; //007;//std::numeric_limits<double>::epsilon();
	const float FPMIN= 9.86076e-031f; //032;//std::numeric_limits<double>::min()/EPS;
	float4 an,b,c,d,del,h;
	//float gln=gammln(a);
	//R: only 31 or 31.5 values for a so I put precomputed gammln(a) here.
	float gln=(a==31.f)?74.658233642578125f:76.371200561523437f;

	b=x+(float4)(1.0f-a);
	c=(float4)(1.0f/FPMIN);
	d=(float4)1.0f/b;
	h=d;
	for (int i=1;i<=ITMAX;++i) {
		an = -i*((float)i-a);
		b += (float4)2.0f;
		d=an*d+b;
		if (fabs(d.x)<FPMIN) d.x=FPMIN;
		if (fabs(d.y)<FPMIN) d.y=FPMIN;
		if (fabs(d.z)<FPMIN) d.z=FPMIN;
		if (fabs(d.w)<FPMIN) d.w=FPMIN;
		c=b+an/c;
		if (fabs(c.x)<FPMIN) c.x=FPMIN;
		if (fabs(c.y)<FPMIN) c.y=FPMIN;
		if (fabs(c.z)<FPMIN) c.z=FPMIN;
		if (fabs(c.w)<FPMIN) c.w=FPMIN;
		d=(float4)1.0f/d;
		del=d*c;
		h*=del;
		if (fabs(del.x-1.0f)<EPS && fabs(del.y-1.0f)<EPS && fabs(del.z-1.0f)<EPS && fabs(del.w-1.0f)<EPS) break;
	}

	return (log(h)-x+a*log(x)-(float4)gln);
}

float4 calc_GaussFit_score(float4 chisqr, float4 null_chisqr,float score_offset) {
	float gauss_bins = 64.0f;
	float gauss_dof = (gauss_bins-2.0f) * 0.5f;
	float null_dof = (gauss_bins-1.0f) * 0.5f;
	gauss_bins *= 0.5f;
	return  score_offset +
		lcgf(gauss_dof,max(chisqr*gauss_bins,(float4)(gauss_dof+1.0f)))
		//-cudaAcc_lcgf(gauss_dof,cudaAcc_GaussFit_settings.gauss_chi_sq_thresh*gauss_bins) // <- always the same result
		-lcgf(null_dof,max(null_chisqr*gauss_bins,null_dof+1.0f));
		//+cudaAcc_lcgf(null_dof,cudaAcc_GaussFit_settings.gauss_null_chi_sq_thresh*gauss_bins); // <- always the same result	
}
#if 0
float4 calc_GaussFit_score_cached(float4 chisqr, float4 null_chisqr,float score_offset,
								  __constant float* gauss, __constant float* null) {
//R: TODO: if validation will fail: increase cache size, implement software linear interpolation
	float4 score=(float4)score_offset;
	int offset;
	offset=(int)max(0.f,min(8191.f,(chisqr.x-1)/10.f*8192.f));
	score.x+=gauss[offset];
	offset=(int)max(0.f,min(8191.f,(null_chisqr.x-1)/10.f*8192.f));
	score.x+=null[offset];
	offset=(int)max(0.f,min(8191.f,(chisqr.y-1)/10.f*8192.f));
	score.y+=gauss[offset];
	offset=(int)max(0.f,min(8191.f,(null_chisqr.y-1)/10.f*8192.f));
	score.y+=null[offset];
	offset=(int)max(0.f,min(8191.f,(chisqr.z-1)/10.f*8192.f));
	score.z+=gauss[offset];
	offset=(int)max(0.f,min(8191.f,(null_chisqr.z-1)/10.f*8192.f));
	score.z+=null[offset];
	offset=(int)max(0.f,min(8191.f,(chisqr.w-1)/10.f*8192.f));
	score.w+=gauss[offset];
	offset=(int)max(0.f,min(8191.f,(null_chisqr.w-1)/10.f*8192.f));
	score.w+=null[offset];

	return  score;
}
#endif
float GetTrueMean2_fl(__global float* fp_PoTPrefixSum, int ul_TOffset, int ul_ExcludeLen, int ul_FftLength,int tid) {
	// TrueMean is the mean power of the data set minus all power
	// out to ExcludeLen from our current TOffset.
	int i_start, i_lim;
	float f_ExcludePower;

	// take care that we do not add to exclude power beyond PoT bounds!
	i_start = max(ul_TOffset - ul_ExcludeLen, 0) - 1;
	i_lim = min(ul_TOffset + ul_ExcludeLen + 1, 64) - 1;

	f_ExcludePower = fp_PoTPrefixSum[i_lim * (ul_FftLength)+tid];
	if (i_start >= 0)
		f_ExcludePower -= fp_PoTPrefixSum[i_start * (ul_FftLength)+tid]; 

	return(((float)(64.0f) - f_ExcludePower) / (float)(64 - (i_lim - i_start)));
}
float GetPeak_fl(__global float* fp_PoT, int ul_TOffset, int ul_HalfSumLength, 
					   float f_MeanPower, float f_PeakScaleFactor, int ul_FftLength,int tid,__constant float* f_weight) {
	// Peak power is calculated as the weighted
	// sum of all powers within ul_HalfSumLength
	// of the assumed gaussian peak.
	// The weights are given by the gaussian function itself.
	// BUG WATCH : for the f_PeakScaleFactor to work,
	// ul_HalfSumLength *must* be set to sigma.

	int i;
	float f_sum=(float)0.0f;

	// Find a weighted sum
	for (i = ul_TOffset - ul_HalfSumLength; i <= ul_TOffset + ul_HalfSumLength; i++) {
		f_sum += (fp_PoT[i * (ul_FftLength)+tid] - f_MeanPower) * f_weight[abs(i-ul_TOffset)];
	}
	return(f_sum * f_PeakScaleFactor);
}

float GetChiSq_fl(__global float* fp_PoT, int ul_FftLength,int ul_TOffset, 
				float f_PeakPower, float f_MeanPower, float* xsq_null,int tid,__constant float* f_weight) {
	// We calculate our assumed gaussian powers
	// on the fly as we try to fit them to the
	// actual powers at each point along the PoT.

	float f_ChiSq = (float)0.0f, f_null_hyp=(float)0.0f;
	float rebin = 1024*1024 / ul_FftLength / 64;

	float recip_MeanPower = (float)1.0f / f_MeanPower;	
	for (int i = 0; i < 64; ++i) {				
		float f_PredictedPower = f_MeanPower + f_PeakPower * (float)f_weight[abs(i - ul_TOffset)];		
		f_PredictedPower *= recip_MeanPower;

		// ChiSq in this realm is:
		//  sum[0:i]( (observed power - expected power)^2 / expected variance )
		// The power of a signal is:
		//  power = (noise + signal)^2 = noise^2 + signal^2 + 2*noise*signal
		// With mean power normalization, noise becomes 1, leaving:
		//  power = signal^2 +or- 2*signal + 1
		float noise=2.0f*sqrt(max(f_PredictedPower,1.0f)-1.0f)+1.0f;
		float recip_noise = (float)rebin / noise;

		float PoTval = fp_PoT[i * (ul_FftLength)+tid] * recip_MeanPower;
		f_ChiSq += (recip_noise*(PoTval - f_PredictedPower)*(PoTval - f_PredictedPower));
		f_null_hyp+= (recip_noise*(PoTval - 1.0f)*(PoTval - 1.0f));
	}

	f_ChiSq/=(float)(64.0f);
	f_null_hyp/=(float)(64.0f);

	(*xsq_null)=f_null_hyp;
	return f_ChiSq;
}
float lcgf_fl(float a, float x) {
	const float EPS= 1.19209e-006f; //007;//std::numeric_limits<double>::epsilon();
	const float FPMIN= 9.86076e-031f; //032;//std::numeric_limits<double>::min()/EPS;
	float an,b,c,d,del,h;
	//float gln=gammln(a);
	//R: only 31 or 31.5 values for a so I put precomputed gammln(a) here.
	float gln=(a==31.f)?74.658233642578125f:76.371200561523437f;

	b=x+(float)(1.0f-a);
	c=(float)(1.0f/FPMIN);
	d=(float)1.0f/b;
	h=d;
	for (int i=1;i<=ITMAX;++i) {
		an = -i*((float)i-a);
		b += (float)2.0f;
		d=an*d+b;
		if (fabs(d)<FPMIN) d=FPMIN;
		c=b+an/c;
		if (fabs(c)<FPMIN) c=FPMIN;
		d=(float)1.0f/d;
		del=d*c;
		h*=del;
		if (fabs(del-1.0f)<EPS) break;
	}

	return (log(h)-x+a*log(x)-(float)gln);
}

float calc_GaussFit_score_fl(float chisqr, float null_chisqr,float score_offset) {
	float gauss_bins = 64.0f;
	float gauss_dof = (gauss_bins-2.0f) * 0.5f;
	float null_dof = (gauss_bins-1.0f) * 0.5f;
	gauss_bins *= 0.5f;
	return  score_offset +
		lcgf_fl(gauss_dof,max(chisqr*gauss_bins,(float)(gauss_dof+1.0f)))
		//-cudaAcc_lcgf(gauss_dof,cudaAcc_GaussFit_settings.gauss_chi_sq_thresh*gauss_bins) // <- always the same result
		-lcgf_fl(null_dof,max(null_chisqr*gauss_bins,null_dof+1.0f));
		//+cudaAcc_lcgf(null_dof,cudaAcc_GaussFit_settings.gauss_null_chi_sq_thresh*gauss_bins); // <- always the same result	
}

#if 0 //R: Vectorized (float4) versions
__kernel __attribute__((vec_type_hint(float4)))
void GaussFit_kernel_cl(__global float4* PoT, __global float4* PoTPrefixSum,int ul_FftLength, float best_gauss_score, 
								 __constant ocl_GaussFit_t* settings,__constant float* f_weight,
								 __global uint4* result_flag, __global float4* GaussFitResults,__global int4* debug) {
//R: each workitem handles 1 dot from 4 PoTs at once.
    int tid =  get_global_id(0);//R: from 0 to fftlen/4    
	int ul_TOffset =  get_global_id(1) + settings->GaussTOffsetStart;//R: something less than 64
	float4 f_null_hyp;

	int iSigma = settings->iSigma;

	float4 	f_TrueMean,
		f_ChiSq,
		f_PeakPower;

	// slide dynamic gaussian across the Power Of Time array

	// TrueMean is the mean power of the data set minus all power
	// out to 2 sigma from our current TOffset.	
	f_TrueMean = GetTrueMean2(
		PoTPrefixSum,
		ul_TOffset,
		2 * iSigma,
		ul_FftLength,tid
		);

	f_PeakPower = GetPeak(
		PoT,
		ul_TOffset,
		iSigma,
		f_TrueMean,
		settings->PeakScaleFactor,
		ul_FftLength,tid,f_weight
		);
	int4 res;
#if 1
//R: cause each workitem deals with 4 PoTs preliminary exit possible only when all 4 have no peak
	// worth looking at ?
	res=(-1)*(f_PeakPower  < settings->GaussPeakPowerThresh3*f_TrueMean);
	//res&=1;
	debug[ul_TOffset * ul_FftLength + tid]=res;
	if (res.x&res.y&res.z&res.w) {
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+1] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+2] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+3] = (float4)0.0f;
		result_flag[0].w=1;
		return;
	}
#endif
	// look at it - try to fit  
	f_ChiSq = GetChiSq(
		PoT,
		ul_FftLength,
		ul_TOffset,
		f_PeakPower,
		f_TrueMean,        
		&f_null_hyp,tid,f_weight
		);
	float4 tmp;

	//R: additional check to see if score needed or not
/*
	res=(f_ChiSq <=  (float4)settings->gauss_chi_sq_thresh);
	if( ((f_ChiSq.x <=  settings->GaussChiSqThresh) && (f_null_hyp.x >= settings->gauss_null_chi_sq_thresh))||
		((f_ChiSq.y <=  settings->GaussChiSqThresh) && (f_null_hyp.y >= settings->gauss_null_chi_sq_thresh)) ||
		((f_ChiSq.z <=  settings->GaussChiSqThresh) && (f_null_hyp.z >= settings->gauss_null_chi_sq_thresh)) ||
		((f_ChiSq.w <=  settings->GaussChiSqThresh) && (f_null_hyp.w >= settings->gauss_null_chi_sq_thresh))
	){
		result_flag[0].z=1;
	}
*/
	float4 score = calc_GaussFit_score(f_ChiSq, f_null_hyp,settings->score_offset);
	
	if (((f_ChiSq.x <=  settings->gauss_chi_sq_thresh) && (score.x > best_gauss_score)) 
		|| ((f_ChiSq.x <=  settings->GaussChiSqThresh) && (f_null_hyp.x >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0].x = 1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean.x;
		tmp.y=f_PeakPower.x;
		tmp.z=f_ChiSq.x;
		tmp.w=f_null_hyp.x;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)] = tmp;
	if (((f_ChiSq.y <=  settings->gauss_chi_sq_thresh) && (score.y > best_gauss_score)) 
		|| ((f_ChiSq.y <=  settings->GaussChiSqThresh) && (f_null_hyp.y >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0].x = 1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean.y;
		tmp.y=f_PeakPower.y;
		tmp.z=f_ChiSq.y;
		tmp.w=f_null_hyp.y;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+1] = tmp;
	if (((f_ChiSq.z <=  settings->gauss_chi_sq_thresh) && (score.z > best_gauss_score)) 
		|| ((f_ChiSq.z <=  settings->GaussChiSqThresh) && (f_null_hyp.z >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0].x = 1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean.z;
		tmp.y=f_PeakPower.z;
		tmp.z=f_ChiSq.z;
		tmp.w=f_null_hyp.z;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+2] = tmp;
	if (((f_ChiSq.w <=  settings->gauss_chi_sq_thresh) && (score.w > best_gauss_score)) 
		|| ((f_ChiSq.w <=  settings->GaussChiSqThresh) && (f_null_hyp.w >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0].x = 1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean.w;
		tmp.y=f_PeakPower.w;
		tmp.z=f_ChiSq.w;
		tmp.w=f_null_hyp.w;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+3] = tmp;
} // End of gaussfit()
__kernel __attribute__((vec_type_hint(float4)))
void GaussFit_no_best_kernel_cl(__global float4* PoT, __global float4* PoTPrefixSum,int ul_FftLength,// float best_gauss_score, 
								 __constant ocl_GaussFit_t* settings,__constant float* f_weight,
								 __global uint4* result_flag, __global float4* GaussFitResults) {
//R: each workitem handles 1 dot from 4 PoTs at once.
    int tid =  get_global_id(0);//R: from 0 to fftlen/4    
	int ul_TOffset =  get_global_id(1) + settings->GaussTOffsetStart;//R: something less than 64
	float4 f_null_hyp;

	int iSigma = settings->iSigma;

	float4 	f_TrueMean,
		f_ChiSq,
		f_PeakPower;

	// slide dynamic gaussian across the Power Of Time array

	// TrueMean is the mean power of the data set minus all power
	// out to 2 sigma from our current TOffset.	
	f_TrueMean = GetTrueMean2(
		PoTPrefixSum,
		ul_TOffset,
		2 * iSigma,
		ul_FftLength,tid
		);

	f_PeakPower = GetPeak(
		PoT,
		ul_TOffset,
		iSigma,
		f_TrueMean,
		settings->PeakScaleFactor,
		ul_FftLength,tid,f_weight
		);
#if 0
//R: cause each workitem deals with 4 PoTs preliminary exit possible only when all 4 have no peak
	// worth looking at ?
	if (f_PeakPower.x / f_TrueMean.x < settings->GaussPeakPowerThresh3 && f_PeakPower.y / f_TrueMean.y < settings->GaussPeakPowerThresh3 &&
		f_PeakPower.z / f_TrueMean.z < settings->GaussPeakPowerThresh3 && f_PeakPower.w / f_TrueMean.w < settings->GaussPeakPowerThresh3) {
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+1] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+2] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+3] = (float4)0.0f;
		return;
	}
#endif
	// look at it - try to fit  
	f_ChiSq = GetChiSq(
		PoT,
		ul_FftLength,
		ul_TOffset,
		f_PeakPower,
		f_TrueMean,        
		&f_null_hyp,tid,f_weight
		);


	//float4 score = calc_GaussFit_score(f_ChiSq, f_null_hyp,settings->score_offset);
	float4 tmp;
	if ( ((f_ChiSq.x <=  settings->GaussChiSqThresh) && (f_null_hyp.x >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0] = (uint4)1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean.x;
		tmp.y=f_PeakPower.x;
		tmp.z=f_ChiSq.x;
		tmp.w=f_null_hyp.x;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)] = tmp;
	if ( ((f_ChiSq.y <=  settings->GaussChiSqThresh) && (f_null_hyp.y >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0] = (uint4)1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean.y;
		tmp.y=f_PeakPower.y;
		tmp.z=f_ChiSq.y;
		tmp.w=f_null_hyp.y;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+1] = tmp;
	if ( ((f_ChiSq.z <=  settings->GaussChiSqThresh) && (f_null_hyp.z >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0] = (uint4)1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean.z;
		tmp.y=f_PeakPower.z;
		tmp.z=f_ChiSq.z;
		tmp.w=f_null_hyp.z;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+2] = tmp;
	if ( ((f_ChiSq.w <=  settings->GaussChiSqThresh) && (f_null_hyp.w >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0] = (uint4)1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean.w;
		tmp.y=f_PeakPower.w;
		tmp.z=f_ChiSq.w;
		tmp.w=f_null_hyp.w;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+3] = tmp;
} // End of gaussfit()

#elif USE_OPENCL_NV || USE_OPENCL_HD5xxx //R: this kernel uses local memory, use it on shared memory-enabled GPUs
#if 0
__constant sampler_t read_sampler = CLK_NORMALIZED_COORDS_TRUE | CLK_ADDRESS_CLAMP_TO_EDGE | CLK_FILTER_LINEAR; 
float calc_GaussFit_score_cached(float chisqr, float null_chisqr,float score_offset,image2d_t gauss_cache,image2d_t null_cache) { // <- gauss_pot_length constant across whole package
	float chisqr_cache = (chisqr - 1.0f) / 10.f; //R: normalized coords clamped to the edge
	float null_chisqr_cache = (null_chisqr - 1.0f) / 10.f;
	return score_offset+ 
		(read_imagef(gauss_cache, read_sampler, (int2)(chisqr_cache, 1))).x+
		(read_imagef(null_cache, read_sampler, (int2)(chisqr_cache, 1))).x;
}
#endif
float GetPeak_local(__local float* fp_PoT, int ul_TOffset, int ul_HalfSumLength, 
					   float f_MeanPower, float f_PeakScaleFactor, int ul_FftLength,__local float* f_weight) {
	// Peak power is calculated as the weighted
	// sum of all powers within ul_HalfSumLength
	// of the assumed gaussian peak.
	// The weights are given by the gaussian function itself.
	// BUG WATCH : for the f_PeakScaleFactor to work,
	// ul_HalfSumLength *must* be set to sigma.

	int i;
	float f_sum=(float)0.0f;

	// Find a weighted sum
	for (i = ul_TOffset - ul_HalfSumLength; i <= ul_TOffset + ul_HalfSumLength; i++) {
		f_sum += (fp_PoT[i*get_local_size(0)+get_local_id(0)] - f_MeanPower) * f_weight[abs(i-ul_TOffset)];
	}
	return(f_sum * f_PeakScaleFactor);
}
float GetChiSq_local(__local float* fp_PoT, int ul_FftLength,int ul_TOffset, 
				float f_PeakPower, float f_MeanPower, float* xsq_null,__local float* f_weight) {
	// We calculate our assumed gaussian powers
	// on the fly as we try to fit them to the
	// actual powers at each point along the PoT.

	float f_ChiSq = (float)0.0f, f_null_hyp=(float)0.0f;
	float rebin = 1024*1024 / ul_FftLength / 64;

	float recip_MeanPower = (float)1.0f / f_MeanPower;	
	for (int i = 0; i < 64; ++i) {				
		float f_PredictedPower = f_MeanPower + f_PeakPower * (float)f_weight[abs(i - ul_TOffset)];		
		f_PredictedPower *= recip_MeanPower;

		// ChiSq in this realm is:
		//  sum[0:i]( (observed power - expected power)^2 / expected variance )
		// The power of a signal is:
		//  power = (noise + signal)^2 = noise^2 + signal^2 + 2*noise*signal
		// With mean power normalization, noise becomes 1, leaving:
		//  power = signal^2 +or- 2*signal + 1
		float noise=2.0f*sqrt(max(f_PredictedPower,1.0f)-1.0f)+1.0f;
		float recip_noise = (float)rebin / noise;

		float PoTval = fp_PoT[i * get_local_size(0)+get_local_id(0)] * recip_MeanPower;
		f_ChiSq += (recip_noise*(PoTval - f_PredictedPower)*(PoTval - f_PredictedPower));
		f_null_hyp+= (recip_noise*(PoTval - 1.0f)*(PoTval - 1.0f));
	}

	f_ChiSq/=(float)(64.0f);
	f_null_hyp/=(float)(64.0f);

	(*xsq_null)=f_null_hyp;
	return f_ChiSq;
}

__kernel void GaussFit_kernel_cl(__global float* PoT, __global float* PoTPrefixSum,int ul_FftLength, float best_gauss_score, 
								 __constant ocl_GaussFit_t* settings,__constant float* f_weight,
								 __global uint4* result_flag, __global float4* GaussFitResults,__global int* debug,
								 __local float* p//,image2d_t gauss_cache,image2d_t null_cache
								 ) {
    int tid =  get_global_id(0);//R: from 0 to fftlen    
	int ul_TOffset =  get_global_id(1) + settings->GaussTOffsetStart;//R: something less than 64
	for(int i=0;i<(64/get_local_size(1));i++){ 
		p[(get_local_size(1)*i+get_local_id(1))*get_local_size(0)+get_local_id(0)]=PoT[tid+(get_local_size(1)*i+get_local_id(1))*(ul_FftLength)];
	}//R: fetch get_local_size(0) PoTs into local memory
	__local float l_weight[64];
	int ltid=(get_local_size(0)*get_local_id(1)+get_local_id(0));
	(ltid<64)?(l_weight[ltid]=f_weight[ltid]):false;//R: fetch weights array into local memory too
	barrier(CLK_LOCAL_MEM_FENCE);
//R: now all accesses to PoT will go into local memory.
if(ul_TOffset>=settings->GaussTOffsetStop) return;//R: we outside of possible gaussian
	float f_null_hyp;

	int iSigma = settings->iSigma;

	float 	f_TrueMean,
		f_ChiSq,
		f_PeakPower;

	// slide dynamic gaussian across the Power Of Time array

	// TrueMean is the mean power of the data set minus all power
	// out to 2 sigma from our current TOffset.	
	f_TrueMean = GetTrueMean2_fl(
		PoTPrefixSum,
		ul_TOffset,
		2 * iSigma,
		ul_FftLength,tid
		);

	f_PeakPower = GetPeak_local(
		p,
		ul_TOffset,
		iSigma,
		f_TrueMean,
		settings->PeakScaleFactor,
		ul_FftLength,l_weight
		);
#if 1
	// worth looking at ?
		int res=(f_PeakPower  < settings->GaussPeakPowerThresh3*f_TrueMean);
	//res&=1;
	//debug[ul_TOffset * ul_FftLength + tid]=res;
	if (res) {
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = (float4)0.0f;
		//result_flag[0].w=1;
		return;
	}
#endif
	// look at it - try to fit  
	f_ChiSq = GetChiSq_local(
		p,
		ul_FftLength,
		ul_TOffset,
		f_PeakPower,
		f_TrueMean,        
		&f_null_hyp,l_weight
		);
	float4 tmp;

	//R: additional check to see if score needed or not
/*
	res=(f_ChiSq <=  (float4)settings->gauss_chi_sq_thresh);
	if( ((f_ChiSq.x <=  settings->GaussChiSqThresh) && (f_null_hyp.x >= settings->gauss_null_chi_sq_thresh))||
		((f_ChiSq.y <=  settings->GaussChiSqThresh) && (f_null_hyp.y >= settings->gauss_null_chi_sq_thresh)) ||
		((f_ChiSq.z <=  settings->GaussChiSqThresh) && (f_null_hyp.z >= settings->gauss_null_chi_sq_thresh)) ||
		((f_ChiSq.w <=  settings->GaussChiSqThresh) && (f_null_hyp.w >= settings->gauss_null_chi_sq_thresh))
	){
		result_flag[0].z=1;
	}
*/
#if 0
	float score = calc_GaussFit_score_cached(f_ChiSq, f_null_hyp,settings->score_offset,gauss_cache,null_cache);
#else
	float score = calc_GaussFit_score_fl(f_ChiSq, f_null_hyp,settings->score_offset);
#endif
	if (((f_ChiSq <=  settings->gauss_chi_sq_thresh) && (score > best_gauss_score)) 
		|| ((f_ChiSq <=  settings->GaussChiSqThresh) && (f_null_hyp >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0].x = 1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean;
		tmp.y=f_PeakPower;
		tmp.z=f_ChiSq;
		tmp.w=f_null_hyp;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = tmp;
} // End of gaussfit()

__kernel void GaussFit_no_best_kernel_cl(__global float* PoT, __global float* PoTPrefixSum,int ul_FftLength, 
								 __constant ocl_GaussFit_t* settings,__constant float* f_weight,
								 __global uint4* result_flag, __global float4* GaussFitResults,__local float* p) {
    int tid =  get_global_id(0);//R: from 0 to fftlen    
	int ul_TOffset =  get_global_id(1) + settings->GaussTOffsetStart;//R: something less than 64
	for(int i=0;i<(64/get_local_size(1));i++){ 
		p[(get_local_size(1)*i+get_local_id(1))*get_local_size(0)+get_local_id(0)]=PoT[tid+(get_local_size(1)*i+get_local_id(1))*(ul_FftLength)];
	}//R: fetch get_local_size(0) PoTs into local memory
	__local float l_weight[64];
	int ltid=(get_local_size(0)*get_local_id(1)+get_local_id(0));
	(ltid<64)?(l_weight[ltid]=f_weight[ltid]):false;//R: fetch weights array into local memory too
	barrier(CLK_LOCAL_MEM_FENCE);
//R: now all accesses to PoT will go into local memory.
if(ul_TOffset>=settings->GaussTOffsetStop) return;//R: we outside of possible gaussian
	float f_null_hyp;

	int iSigma = settings->iSigma;

	float 	f_TrueMean,
		f_ChiSq,
		f_PeakPower;

	// slide dynamic gaussian across the Power Of Time array

	// TrueMean is the mean power of the data set minus all power
	// out to 2 sigma from our current TOffset.	
	f_TrueMean = GetTrueMean2_fl(
		PoTPrefixSum,
		ul_TOffset,
		2 * iSigma,
		ul_FftLength,tid
		);

	f_PeakPower = GetPeak_local(
		p,
		ul_TOffset,
		iSigma,
		f_TrueMean,
		settings->PeakScaleFactor,
		ul_FftLength,l_weight
		);
#if 1
	// worth looking at ?
		int res=(f_PeakPower  < settings->GaussPeakPowerThresh3*f_TrueMean);
	//res&=1;
	//debug[ul_TOffset * ul_FftLength + tid]=res;
	if (res) {
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = (float4)0.0f;
		//result_flag[0].w=1;
		return;
	}
#endif
	// look at it - try to fit  
	f_ChiSq = GetChiSq_local(
		p,
		ul_FftLength,
		ul_TOffset,
		f_PeakPower,
		f_TrueMean,        
		&f_null_hyp,l_weight
		);
	float4 tmp;

	if ( ((f_ChiSq <=  settings->GaussChiSqThresh) && (f_null_hyp >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0].x = 1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean;
		tmp.y=f_PeakPower;
		tmp.z=f_ChiSq;
		tmp.w=f_null_hyp;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = tmp;
} // End of gaussfit()


#else //R: ATI HD4xxx kernel (non vectorized but with pre-exit enabled)
__kernel void GaussFit_kernel_cl(__global float* PoT, __global float* PoTPrefixSum,int ul_FftLength, float best_gauss_score, 
								 __constant ocl_GaussFit_t* settings,__constant float* f_weight,
								 __global uint4* result_flag, __global float4* GaussFitResults,__global int* debug) {
    int tid =  get_global_id(0);//R: from 0 to fftlen    
	int ul_TOffset =  get_global_id(1) + settings->GaussTOffsetStart;//R: something less than 64
	if(ul_TOffset>=settings->GaussTOffsetStop) return;//R: we outside of possible gaussian
	float f_null_hyp;

	int iSigma = settings->iSigma;

	float 	f_TrueMean,
		f_ChiSq,
		f_PeakPower;

	// slide dynamic gaussian across the Power Of Time array

	// TrueMean is the mean power of the data set minus all power
	// out to 2 sigma from our current TOffset.	
	f_TrueMean = GetTrueMean2_fl(
		PoTPrefixSum,
		ul_TOffset,
		2 * iSigma,
		ul_FftLength,tid
		);

	f_PeakPower = GetPeak_fl(
		PoT,
		ul_TOffset,
		iSigma,
		f_TrueMean,
		settings->PeakScaleFactor,
		ul_FftLength,tid,f_weight
		);
#if 1
	int res;
	// worth looking at ?
	res=(f_PeakPower  < settings->GaussPeakPowerThresh3*f_TrueMean);
	//res&=1;
	//debug[ul_TOffset * ul_FftLength + tid]=res;
	if (res) {
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = (float4)0.0f;
		//result_flag[0].w=1;
		return;
	}
#endif
	// look at it - try to fit  
	f_ChiSq = GetChiSq_fl(
		PoT,
		ul_FftLength,
		ul_TOffset,
		f_PeakPower,
		f_TrueMean,        
		&f_null_hyp,tid,f_weight
		);
	float4 tmp;

	//R: additional check to see if score needed or not
/*
	res=(f_ChiSq <=  (float4)settings->gauss_chi_sq_thresh);
	if( ((f_ChiSq.x <=  settings->GaussChiSqThresh) && (f_null_hyp.x >= settings->gauss_null_chi_sq_thresh))||
		((f_ChiSq.y <=  settings->GaussChiSqThresh) && (f_null_hyp.y >= settings->gauss_null_chi_sq_thresh)) ||
		((f_ChiSq.z <=  settings->GaussChiSqThresh) && (f_null_hyp.z >= settings->gauss_null_chi_sq_thresh)) ||
		((f_ChiSq.w <=  settings->GaussChiSqThresh) && (f_null_hyp.w >= settings->gauss_null_chi_sq_thresh))
	){
		result_flag[0].z=1;
	}
*/
	float score = calc_GaussFit_score_fl(f_ChiSq, f_null_hyp,settings->score_offset);
	
	if (((f_ChiSq <=  settings->gauss_chi_sq_thresh) && (score > best_gauss_score)) 
		|| ((f_ChiSq <=  settings->GaussChiSqThresh) && (f_null_hyp >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0].x = 1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean;
		tmp.y=f_PeakPower;
		tmp.z=f_ChiSq;
		tmp.w=f_null_hyp;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = tmp;
} // End of gaussfit()


__kernel void GaussFit_no_best_kernel_cl(__global float* PoT, __global float* PoTPrefixSum,int ul_FftLength, 
								 __constant ocl_GaussFit_t* settings,__constant float* f_weight,
								 __global uint4* result_flag, __global float4* GaussFitResults) {
    int tid =  get_global_id(0);//R: from 0 to fftlen    
	int ul_TOffset =  get_global_id(1) + settings->GaussTOffsetStart;//R: something less than 64
	if(ul_TOffset>=settings->GaussTOffsetStop) return;//R: we outside of possible gaussian
	float f_null_hyp;

	int iSigma = settings->iSigma;

	float 	f_TrueMean,
		f_ChiSq,
		f_PeakPower;

	// slide dynamic gaussian across the Power Of Time array

	// TrueMean is the mean power of the data set minus all power
	// out to 2 sigma from our current TOffset.	
	f_TrueMean = GetTrueMean2_fl(
		PoTPrefixSum,
		ul_TOffset,
		2 * iSigma,
		ul_FftLength,tid
		);

	f_PeakPower = GetPeak_fl(
		PoT,
		ul_TOffset,
		iSigma,
		f_TrueMean,
		settings->PeakScaleFactor,
		ul_FftLength,tid,f_weight
		);
#if 1
	int res;
	// worth looking at ?
	res=(f_PeakPower  < settings->GaussPeakPowerThresh3*f_TrueMean);
	//res&=1;
	//debug[ul_TOffset * ul_FftLength + tid]=res;
	if (res) {
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = (float4)0.0f;
		//result_flag[0].w=1;
		return;
	}
#endif
	// look at it - try to fit  
	f_ChiSq = GetChiSq_fl(
		PoT,
		ul_FftLength,
		ul_TOffset,
		f_PeakPower,
		f_TrueMean,        
		&f_null_hyp,tid,f_weight
		);
	float4 tmp;
	if ( ((f_ChiSq <=  settings->GaussChiSqThresh) && (f_null_hyp >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0].x = 1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean;
		tmp.y=f_PeakPower;
		tmp.z=f_ChiSq;
		tmp.w=f_null_hyp;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = tmp;
} // End of gaussfit()



#endif

#if 0
__kernel __attribute__((vec_type_hint(float4)))
void GaussFit_kernel_cached_lcgf_cl(__global float4* PoT, __global float4* PoTPrefixSum,int ul_FftLength, float best_gauss_score, 
								 __constant ocl_GaussFit_t* settings,__constant float* f_weight,
								 __global uint4* result_flag, __global float4* GaussFitResults,
								 __constant float* gauss, __constant float* null){
//R: each workitem handles 1 dot from 4 PoTs at once.
    int tid =  get_global_id(0);//R: from 0 to fftlen/4    
	int ul_TOffset =  get_global_id(1) + settings->GaussTOffsetStart;//R: something less than 64
	float4 f_null_hyp;

	int iSigma = settings->iSigma;

	float4 	f_TrueMean,
		f_ChiSq,
		f_PeakPower;

	// slide dynamic gaussian across the Power Of Time array

	// TrueMean is the mean power of the data set minus all power
	// out to 2 sigma from our current TOffset.	
	f_TrueMean = GetTrueMean2(
		PoTPrefixSum,
		ul_TOffset,
		2 * iSigma,
		ul_FftLength,tid
		);

	f_PeakPower = GetPeak(
		PoT,
		ul_TOffset,
		iSigma,
		f_TrueMean,
		settings->PeakScaleFactor,
		ul_FftLength,tid,f_weight
		);
	float4 tmp;
	uint4 res;
#if 1
//R: cause each workitem deals with 4 PoTs preliminary exit possible only when all 4 have no peak
	// worth looking at ?
	res=(f_PeakPower  < settings->GaussPeakPowerThresh3*f_TrueMean);
	if (res.x&res.y&res.z&res.w) {
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+1] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+2] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+3] = (float4)0.0f;
		result_flag[0].w=1;
		return;
	}
#endif
	// look at it - try to fit  
	f_ChiSq = GetChiSq(
		PoT,
		ul_FftLength,
		ul_TOffset,
		f_PeakPower,
		f_TrueMean,        
		&f_null_hyp,tid,f_weight
		);

	//R: additional check to see if score needed or not
	tmp=(-1.f)*(float4)(f_ChiSq <=  (float4)settings->gauss_chi_sq_thresh);
	debug[ul_TOffset * ul_FftLength + (tid<<2)+3]=tmp;
	if( ((f_ChiSq.x <=  settings->GaussChiSqThresh) && (f_null_hyp.x >= settings->gauss_null_chi_sq_thresh))||
		((f_ChiSq.y <=  settings->GaussChiSqThresh) && (f_null_hyp.y >= settings->gauss_null_chi_sq_thresh)) ||
		((f_ChiSq.z <=  settings->GaussChiSqThresh) && (f_null_hyp.z >= settings->gauss_null_chi_sq_thresh)) ||
		((f_ChiSq.w <=  settings->GaussChiSqThresh) && (f_null_hyp.w >= settings->gauss_null_chi_sq_thresh))
	){
		result_flag[0].z=1;
	}


	float4 score = calc_GaussFit_score_cached(f_ChiSq, f_null_hyp,settings->score_offset,gauss,null);
	if (((f_ChiSq.x <=  settings->gauss_chi_sq_thresh) && (score.x > best_gauss_score)) 
		|| ((f_ChiSq.x <=  settings->GaussChiSqThresh) && (f_null_hyp.x >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0] = (uint4)1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean.x;
		tmp.y=f_PeakPower.x;
		tmp.z=f_ChiSq.x;
		tmp.w=f_null_hyp.x;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)] = tmp;
	if (((f_ChiSq.y <=  settings->gauss_chi_sq_thresh) && (score.y > best_gauss_score)) 
		|| ((f_ChiSq.y <=  settings->GaussChiSqThresh) && (f_null_hyp.y >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0] = (uint4)1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean.y;
		tmp.y=f_PeakPower.y;
		tmp.z=f_ChiSq.y;
		tmp.w=f_null_hyp.y;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+1] = tmp;
	if (((f_ChiSq.z <=  settings->gauss_chi_sq_thresh) && (score.z > best_gauss_score)) 
		|| ((f_ChiSq.z <=  settings->GaussChiSqThresh) && (f_null_hyp.z >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0] = (uint4)1;//R: results array sh0uld be downloaded to host
		tmp.x=f_TrueMean.z;
		tmp.y=f_PeakPower.z;
		tmp.z=f_ChiSq.z;
		tmp.w=f_null_hyp.z;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+2] = tmp;
	if (((f_ChiSq.w <=  settings->gauss_chi_sq_thresh) && (score.w > best_gauss_score)) 
		|| ((f_ChiSq.w <=  settings->GaussChiSqThresh) && (f_null_hyp.w >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0] = (uint4)1;//R: results array shuld be downloaded to host
		tmp.x=f_TrueMean.w;
		tmp.y=f_PeakPower.w;
		tmp.z=f_ChiSq.w;
		tmp.w=f_null_hyp.w;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+3] = tmp;
} // End of gaussfit()


#endif
__kernel void GaussFit_kernel_PE_cl(__global float4* PoT, __global float4* PoTPrefixSum,int ul_FftLength, float best_gauss_score, 
								 __constant ocl_GaussFit_t* settings,__constant float* f_weight,
								 __global uint4* result_flag, __global float4* GaussFitResults,
								 __constant uint* WeakPeaks) {
//R: each workitem handles 1 dot from 4 PoTs at once.
    int tid =  get_global_id(0);//R: from 0 to fftlen/4    
	if(WeakPeaks[tid]==0)return;//R: preliminary exit if whole 4 PoTs don't have even weak peak
	int ul_TOffset =  get_global_id(1) + settings->GaussTOffsetStart;//R: something less than 64
	float4 f_null_hyp;

	int iSigma = settings->iSigma;

	float4 	f_TrueMean,
		f_ChiSq,
		f_PeakPower;

	// slide dynamic gaussian across the Power Of Time array

	// TrueMean is the mean power of the data set minus all power
	// out to 2 sigma from our current TOffset.	
	f_TrueMean = GetTrueMean2(
		PoTPrefixSum,
		ul_TOffset,
		2 * iSigma,
		ul_FftLength,tid
		);

	f_PeakPower = GetPeak(
		PoT,
		ul_TOffset,
		iSigma,
		f_TrueMean,
		settings->PeakScaleFactor,
		ul_FftLength,tid,f_weight
		);
#if 0
//R: cause each workitem deals with 4 PoTs preliminary exit possible only when all 4 have no peak
	// worth looking at ?
	if (f_PeakPower.x / f_TrueMean.x < settings->GaussPeakPowerThresh3 && f_PeakPower.y / f_TrueMean.y < settings->GaussPeakPowerThresh3 &&
		f_PeakPower.z / f_TrueMean.z < settings->GaussPeakPowerThresh3 && f_PeakPower.w / f_TrueMean.w < settings->GaussPeakPowerThresh3) {
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+1] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+2] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+3] = (float4)0.0f;
		return;
	}
#endif
	// look at it - try to fit  
	f_ChiSq = GetChiSq(
		PoT,
		ul_FftLength,
		ul_TOffset,
		f_PeakPower,
		f_TrueMean,        
		&f_null_hyp,tid,f_weight
		);


	float4 score = calc_GaussFit_score(f_ChiSq, f_null_hyp,settings->score_offset);
	float4 tmp;
	if (((f_ChiSq.x <=  settings->gauss_chi_sq_thresh) && (score.x > best_gauss_score)) 
		|| ((f_ChiSq.x <=  settings->GaussChiSqThresh) && (f_null_hyp.x >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0] = (uint4)1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean.x;
		tmp.y=f_PeakPower.x;
		tmp.z=f_ChiSq.x;
		tmp.w=f_null_hyp.x;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)] = tmp;
	if (((f_ChiSq.y <=  settings->gauss_chi_sq_thresh) && (score.y > best_gauss_score)) 
		|| ((f_ChiSq.y <=  settings->GaussChiSqThresh) && (f_null_hyp.y >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0] = (uint4)1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean.y;
		tmp.y=f_PeakPower.y;
		tmp.z=f_ChiSq.y;
		tmp.w=f_null_hyp.y;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+1] = tmp;
	if (((f_ChiSq.z <=  settings->gauss_chi_sq_thresh) && (score.z > best_gauss_score)) 
		|| ((f_ChiSq.z <=  settings->GaussChiSqThresh) && (f_null_hyp.z >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0] = (uint4)1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean.z;
		tmp.y=f_PeakPower.z;
		tmp.z=f_ChiSq.z;
		tmp.w=f_null_hyp.z;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+2] = tmp;
	if (((f_ChiSq.w <=  settings->gauss_chi_sq_thresh) && (score.w > best_gauss_score)) 
		|| ((f_ChiSq.w <=  settings->GaussChiSqThresh) && (f_null_hyp.w >= settings->gauss_null_chi_sq_thresh))
		) {
		result_flag[0] = (uint4)1;//R: results array should be downloaded to host
		tmp.x=f_TrueMean.w;
		tmp.y=f_PeakPower.w;
		tmp.z=f_ChiSq.w;
		tmp.w=f_null_hyp.w;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+3] = tmp;
} // End of gaussfit()

__kernel __attribute__((work_group_size_hint(1, 1, 1)))
void set_mem_kernel_cl(__global float4* data){
	uint tid=get_global_id(0);
	data[tid]=(float4)(0.0f);
}
#define MAX_BINS_ABOVE_THRESHOLD 8 //R: depends on how long private array can be not spilled into global memory
__kernel void find_triplets_kernel_cl(int ul_FftLength, int len_power, float triplet_thresh_base, int AdvanceBy, int PoTLen,
									  __global float4* power,__global float4* triplet_results,__global int4* result_flag) {
#if 0
//R: this kernel does 4 PoT chunks at once.
//R: cause workitems can write flag in arbitrary order it's not possible to set error code actually (original CUDA code
// missed this fact and tries to return error code from kernel. That is, different variable should be used for setting 
// state flags. In this kernel next convention used:
//If whole flag is zero - there is no errors and no result loading on CPU required, kernel launch was silent.
//If result_flag.x not zero - but all other parts are zero - kernel launch was OK, some triplets found, CPU logging required.
// result_flag.y!=0 - there were too many bins above threshold, CPU reprocessing required
// result_flag.z!=0 -there were too many (>2) triplets found in single PoT chunk, CPU reprocessing required
// result_flag.w!=0 - currently reserved.
	int ul_PoT = get_global_id(0);//R: 4 PoTs at once!
	int y = get_global_id(1);//R: index of offset chunk
	int TOffset = y * AdvanceBy;

	if(TOffset + len_power > PoTLen) {		
		TOffset = PoTLen - len_power;
	}
	// Clear the result array
	triplet_results[(ul_PoT<<2)+ (2*y)*(ul_FftLength)] = (float4)(0.0f);
	triplet_results[(ul_PoT<<2)+ (2*y+1)*(ul_FftLength)] = (float4)(0.0f);
	triplet_results[1+(ul_PoT<<2)+ (2*y)*(ul_FftLength)] = (float4)(0.0f);
	triplet_results[1+(ul_PoT<<2)+ (2*y+1)*(ul_FftLength)] = (float4)(0.0f);
	triplet_results[2+(ul_PoT<<2)+ (2*y)*(ul_FftLength)] = (float4)(0.0f);
	triplet_results[2+(ul_PoT<<2)+ (2*y+1)*(ul_FftLength)] = (float4)(0.0f);
	triplet_results[3+(ul_PoT<<2)+ (2*y)*(ul_FftLength)] = (float4)(0.0f);
	triplet_results[3+(ul_PoT<<2)+ (2*y+1)*(ul_FftLength)] = (float4)(0.0f);
	int4 binsAboveThreshold[MAX_BINS_ABOVE_THRESHOLD],numBinsAboveThreshold=(int4)0;
	int i,n,p,q;
	float4 midpoint,mean_power=0,peak_power,period,triplet_thresh=(float4)triplet_thresh_base,pp;

	/* Get all the bins that are above the threshold, and find the power array mean value */
	for( i=0;i<len_power;i++ ) {
		mean_power += power[ul_PoT+(ul_FftLength>>2)*TOffset];
	}
	mean_power /= (float4)len_power;
	triplet_thresh*=mean_power;

	for( i=0;i<len_power;i++ ) {
		pp=power[ul_PoT+(ul_FftLength>>2)*TOffset];
		if(  pp.x>= triplet_thresh.x ) {
			if (numBinsAboveThreshold.x == MAX_BINS_ABOVE_THRESHOLD) {
				 result_flag[0].y=1; // Reporting Error
				return;
			}
		binsAboveThreshold[numBinsAboveThreshold.x].x = i;
		numBinsAboveThreshold.x++;
		}
		if(  pp.y>= triplet_thresh.y ) {
			if (numBinsAboveThreshold.y == MAX_BINS_ABOVE_THRESHOLD) {
				 result_flag[0].y=1; // Reporting Error
				return;
			}
		binsAboveThreshold[numBinsAboveThreshold.y].y = i;
		numBinsAboveThreshold.y++;
		}
		if(  pp.z>= triplet_thresh.z ) {
			if (numBinsAboveThreshold.z == MAX_BINS_ABOVE_THRESHOLD) {
				 result_flag[0].y=1; // Reporting Error
				return;
			}
		binsAboveThreshold[numBinsAboveThreshold.z].z = i;
		numBinsAboveThreshold.z++;
		}
		if(  pp.w>= triplet_thresh.w ) {
			if (numBinsAboveThreshold.w == MAX_BINS_ABOVE_THRESHOLD) {
				 result_flag[0].y=1; // Reporting Error
				return;
			}
		binsAboveThreshold[numBinsAboveThreshold.w].w = i;
		numBinsAboveThreshold.w++;
		}
	}

	/* Check each bin combination for a triplet */
	if (numBinsAboveThreshold.x>2) { //R: should be rare case and _maybe_ it's worth to read 4 floats there and use only one of them
		int already_reported_flag = 0;
		// TODO: Use Texture for reads from power[], Random Reads (R: it means to use OpenCL image )
		for( i=0;i<numBinsAboveThreshold.x-1;i++ ) {
			for( n=i+2;n<numBinsAboveThreshold.x;n++ ) {
				midpoint.x = (binsAboveThreshold[i].x+binsAboveThreshold[n].x)/2.0f;
				period.x = (float)fabs((binsAboveThreshold[i].x-binsAboveThreshold[n].x)/2.0f);

				/* Get the peak power of this triplet */
				//R: read 4 bins from different PoTs but use only needed one
				pp=power[ul_PoT+(ul_FftLength>>2)*binsAboveThreshold[i].x];
				peak_power.x = pp.x;
				pp=power[ul_PoT+(ul_FftLength>>2)*binsAboveThreshold[n].x];
				peak_power.x = max(peak_power.x,pp.x);

				p = binsAboveThreshold[i].x;
				while( (power[ul_PoT+(ul_FftLength>>2)*p].x >= triplet_thresh.x) && 
					(p <= (int)(floor(midpoint.x))) ) {    /* Check if there is a pulse "off" in between init and midpoint */
					p++;
				}

				q = (int)(floor(midpoint.x))+1;
				while( (power[ul_PoT+(ul_FftLength>>2)*q].x >= triplet_thresh.x) && (q <= binsAboveThreshold[n].x)) {    /* Check if there is a pulse "off" in between midpoint and end */
					q++;
				}

				if( p >= (int)(floor(midpoint.x)) || q >= binsAboveThreshold[n].x) {
					/* if this pulse doesn't have an "off" between all the three spikes, it's dropped */
				} else if( (midpoint.x - floor(midpoint.x)) > 0.1f ) {    /* if it's spread among two bins */
					if( power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.x].x >= triplet_thresh.x ) {
						if (already_reported_flag >= 2)	{
							 result_flag[0].z=1; // Reporting Error, more than 2 results per PoT, redo the calculations on CPU
							return;
						}
						peak_power.x = max(peak_power.x,power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.x].x);
						float4 tmp;
						tmp.x=peak_power.x/mean_power.x;
						tmp.y=mean_power.x;tmp.z=period.x;tmp.w=midpoint.x;
						triplet_results[ul_PoT+ (2*TOffset+already_reported_flag)*(ul_FftLength) + 0/*1 for .y and so on*/]=tmp; 
						result_flag[0].x= 1; // Mark for download <- VERY RARE SITUATION
						already_reported_flag++;
						//ReportTripletEvent( peak_power/mean_power, mean_power, period, midpoint,time_bin, freq_bin, len_power, power, 1 );
					}

					if( power[ul_PoT+(ul_FftLength>>2)*(int)(midpoint.x+1.f)].x >= triplet_thresh.x ) {
						if (already_reported_flag >= 2)	{
							result_flag[0].z=1; // Reporting Error, more than one result per PoT, redo the calculations on CPU
							return;
						}
						peak_power.x = max(peak_power.x,power[ul_PoT+(ul_FftLength>>2)*(int)(midpoint.x+1.f)].x);
						float4 tmp;
						tmp.x=peak_power.x/mean_power.x;
						tmp.y=mean_power.x;tmp.z=period.x;tmp.w=midpoint.x;
						triplet_results[ul_PoT+ (2*TOffset+already_reported_flag)*(ul_FftLength) + 0/*1 for .y and so on*/]=tmp; 
						result_flag[0].x= 1; // Mark for download <- VERY RARE SITUATION
						already_reported_flag++;
						//ReportTripletEvent(peak_power/mean_power, mean_power, period, midpoint,time_bin, freq_bin, len_power, power, 1 );
					}

				} else {            /* otherwise just check the single midpoint bin */
					if( power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.x].x >= triplet_thresh.x ) {
						if (already_reported_flag >= 2)	{
							result_flag[0].z=1; // Reporting Error, more than one result per PoT, redo the calculations on CPU
							return;
						}

						peak_power.x = max(peak_power.x,power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.x].x);

						float4 tmp;
						tmp.x=peak_power.x/mean_power.x;
						tmp.y=mean_power.x;tmp.z=period.x;tmp.w=midpoint.x;
						triplet_results[ul_PoT+ (2*TOffset+already_reported_flag)*(ul_FftLength) + 0/*1 for .y and so on*/]=tmp; 
						result_flag[0].x= 1; // Mark for download <- VERY RARE SITUATION
						already_reported_flag++;
						//ReportTripletEvent( peak_power/mean_power, mean_power, period, midpoint,time_bin, freq_bin, len_power, power, 1 );
					}
				}
			}
		}
	}
	if (numBinsAboveThreshold.y>2) { //R: should be rare case and _maybe_ it's worth to read 4 floats there and use only one of them
		int already_reported_flag = 0;
		// TODO: Use Texture for reads from power[], Random Reads (R: it means to use OpenCL image )
		for( i=0;i<numBinsAboveThreshold.y-1;i++ ) {
			for( n=i+2;n<numBinsAboveThreshold.y;n++ ) {
				midpoint.y = (binsAboveThreshold[i].y+binsAboveThreshold[n].y)/2.0f;
				period.y = (float)fabs((binsAboveThreshold[i].y-binsAboveThreshold[n].y)/2.0f);

				/* Get the peak power of this triplet */
				//R: read 4 bins from different PoTs but use only needed one
				pp=power[ul_PoT+(ul_FftLength>>2)*binsAboveThreshold[i].y];
				peak_power.y = pp.y;
				pp=power[ul_PoT+(ul_FftLength>>2)*binsAboveThreshold[n].y];
				peak_power.y = max(peak_power.y,pp.y);

				p = binsAboveThreshold[i].y;
				while( (power[ul_PoT+(ul_FftLength>>2)*p].y >= triplet_thresh.y) && 
					(p <= (int)(floor(midpoint.y))) ) {    /* Check if there is a pulse "off" in between init and midpoint */
					p++;
				}

				q = (int)(floor(midpoint.y))+1;
				while( (power[ul_PoT+(ul_FftLength>>2)*q].y >= triplet_thresh.y) && (q <= binsAboveThreshold[n].y)) {    /* Check if there is a pulse "off" in between midpoint and end */
					q++;
				}

				if( p >= (int)(floor(midpoint.y)) || q >= binsAboveThreshold[n].y) {
					/* if this pulse doesn't have an "off" between all the three spikes, it's dropped */
				} else if( (midpoint.y - floor(midpoint.y)) > 0.1f ) {    /* if it's spread among two bins */
					if( power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.y].y >= triplet_thresh.y ) {
						if (already_reported_flag >= 2)	{
							 result_flag[0].z=1; // Reporting Error, more than 2 results per PoT, redo the calculations on CPU
							return;
						}
						peak_power.y = max(peak_power.y,power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.y].y);
						float4 tmp;
						tmp.x=peak_power.y/mean_power.y;
						tmp.y=mean_power.y;tmp.z=period.y;tmp.w=midpoint.y;
						triplet_results[ul_PoT+ (2*TOffset+already_reported_flag)*(ul_FftLength) + 1/*1 for .y and so on*/]=tmp; 
						result_flag[0].x= 1; // Mark for download <- VERY RARE SITUATION
						already_reported_flag++;
						//ReportTripletEvent( peak_power/mean_power, mean_power, period, midpoint,time_bin, freq_bin, len_power, power, 1 );
					}

					if( power[ul_PoT+(ul_FftLength>>2)*(int)(midpoint.y+1.f)].y >= triplet_thresh.y ) {
						if (already_reported_flag >= 2)	{
							result_flag[0].z=1; // Reporting Error, more than one result per PoT, redo the calculations on CPU
							return;
						}
						peak_power.y = max(peak_power.y,power[ul_PoT+(ul_FftLength>>2)*(int)(midpoint.y+1.f)].y);
						float4 tmp;
						tmp.x=peak_power.y/mean_power.y;
						tmp.y=mean_power.y;tmp.z=period.y;tmp.w=midpoint.y;
						triplet_results[ul_PoT+ (2*TOffset+already_reported_flag)*(ul_FftLength) + 1/*1 for .y and so on*/]=tmp; 
						result_flag[0].x= 1; // Mark for download <- VERY RARE SITUATION
						already_reported_flag++;
						//ReportTripletEvent(peak_power/mean_power, mean_power, period, midpoint,time_bin, freq_bin, len_power, power, 1 );
					}

				} else {            /* otherwise just check the single midpoint bin */
					if( power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.y].y >= triplet_thresh.y ) {
						if (already_reported_flag >= 2)	{
							result_flag[0].z=1; // Reporting Error, more than one result per PoT, redo the calculations on CPU
							return;
						}

						peak_power.y = max(peak_power.y,power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.y].y);

						float4 tmp;
						tmp.x=peak_power.y/mean_power.y;
						tmp.y=mean_power.y;tmp.z=period.y;tmp.w=midpoint.y;
						triplet_results[ul_PoT+ (2*TOffset+already_reported_flag)*(ul_FftLength) + 1/*1 for .y and so on*/]=tmp; 
						result_flag[0].x= 1; // Mark for download <- VERY RARE SITUATION
						already_reported_flag++;
						//ReportTripletEvent( peak_power/mean_power, mean_power, period, midpoint,time_bin, freq_bin, len_power, power, 1 );
					}
				}
			}
		}
	}
	if (numBinsAboveThreshold.z>2) { //R: should be rare case and _maybe_ it's worth to read 4 floats there and use only one of them
		int already_reported_flag = 0;
		// TODO: Use Texture for reads from power[], Random Reads (R: it means to use OpenCL image )
		for( i=0;i<numBinsAboveThreshold.z-1;i++ ) {
			for( n=i+2;n<numBinsAboveThreshold.z;n++ ) {
				midpoint.z = (binsAboveThreshold[i].z+binsAboveThreshold[n].z)/2.0f;
				period.z = (float)fabs((binsAboveThreshold[i].z-binsAboveThreshold[n].z)/2.0f);

				/* Get the peak power of this triplet */
				//R: read 4 bins from different PoTs but use only needed one
				pp=power[ul_PoT+(ul_FftLength>>2)*binsAboveThreshold[i].z];
				peak_power.z = pp.z;
				pp=power[ul_PoT+(ul_FftLength>>2)*binsAboveThreshold[n].z];
				peak_power.z = max(peak_power.z,pp.z);

				p = binsAboveThreshold[i].z;
				while( (power[ul_PoT+(ul_FftLength>>2)*p].z >= triplet_thresh.z) && 
					(p <= (int)(floor(midpoint.z))) ) {    /* Check if there is a pulse "off" in between init and midpoint */
					p++;
				}

				q = (int)(floor(midpoint.z))+1;
				while( (power[ul_PoT+(ul_FftLength>>2)*q].z >= triplet_thresh.z) && (q <= binsAboveThreshold[n].z)) {    /* Check if there is a pulse "off" in between midpoint and end */
					q++;
				}

				if( p >= (int)(floor(midpoint.z)) || q >= binsAboveThreshold[n].z) {
					/* if this pulse doesn't have an "off" between all the three spikes, it's dropped */
				} else if( (midpoint.z - floor(midpoint.z)) > 0.1f ) {    /* if it's spread among two bins */
					if( power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.z].z >= triplet_thresh.z ) {
						if (already_reported_flag >= 2)	{
							 result_flag[0].z=1; // Reporting Error, more than 2 results per PoT, redo the calculations on CPU
							return;
						}
						peak_power.z = max(peak_power.z,power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.z].z);
						float4 tmp;
						tmp.x=peak_power.z/mean_power.z;
						tmp.y=mean_power.z;tmp.z=period.z;tmp.w=midpoint.z;
						triplet_results[ul_PoT+ (2*TOffset+already_reported_flag)*(ul_FftLength) + 2/*1 for .y and so on*/]=tmp; 
						result_flag[0].x= 1; // Mark for download <- VERY RARE SITUATION
						already_reported_flag++;
						//ReportTripletEvent( peak_power/mean_power, mean_power, period, midpoint,time_bin, freq_bin, len_power, power, 1 );
					}

					if( power[ul_PoT+(ul_FftLength>>2)*(int)(midpoint.z+1.f)].z >= triplet_thresh.z ) {
						if (already_reported_flag >= 2)	{
							result_flag[0].z=1; // Reporting Error, more than one result per PoT, redo the calculations on CPU
							return;
						}
						peak_power.z = max(peak_power.z,power[ul_PoT+(ul_FftLength>>2)*(int)(midpoint.z+1.f)].z);
						float4 tmp;
						tmp.x=peak_power.z/mean_power.z;
						tmp.y=mean_power.z;tmp.z=period.z;tmp.w=midpoint.z;
						triplet_results[ul_PoT+ (2*TOffset+already_reported_flag)*(ul_FftLength) + 2/*1 for .y and so on*/]=tmp; 
						result_flag[0].x= 1; // Mark for download <- VERY RARE SITUATION
						already_reported_flag++;
						//ReportTripletEvent(peak_power/mean_power, mean_power, period, midpoint,time_bin, freq_bin, len_power, power, 1 );
					}

				} else {            /* otherwise just check the single midpoint bin */
					if( power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.z].z >= triplet_thresh.z ) {
						if (already_reported_flag >= 2)	{
							result_flag[0].z=1; // Reporting Error, more than one result per PoT, redo the calculations on CPU
							return;
						}

						peak_power.z = max(peak_power.z,power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.z].z);

						float4 tmp;
						tmp.x=peak_power.z/mean_power.z;
						tmp.y=mean_power.z;tmp.z=period.z;tmp.w=midpoint.z;
						triplet_results[ul_PoT+ (2*TOffset+already_reported_flag)*(ul_FftLength) + 2/*1 for .y and so on*/]=tmp; 
						result_flag[0].x= 1; // Mark for download <- VERY RARE SITUATION
						already_reported_flag++;
						//ReportTripletEvent( peak_power/mean_power, mean_power, period, midpoint,time_bin, freq_bin, len_power, power, 1 );
					}
				}
			}
		}
	}
	if (numBinsAboveThreshold.w>2) { //R: should be rare case and _maybe_ it's worth to read 4 floats there and use only one of them
		int already_reported_flag = 0;
		// TODO: Use Texture for reads from power[], Random Reads (R: it means to use OpenCL image )
		for( i=0;i<numBinsAboveThreshold.w-1;i++ ) {
			for( n=i+2;n<numBinsAboveThreshold.w;n++ ) {
				midpoint.w = (binsAboveThreshold[i].w+binsAboveThreshold[n].w)/2.0f;
				period.w = (float)fabs((binsAboveThreshold[i].w-binsAboveThreshold[n].w)/2.0f);

				/* Get the peak power of this triplet */
				//R: read 4 bins from different PoTs but use only needed one
				pp=power[ul_PoT+(ul_FftLength>>2)*binsAboveThreshold[i].w];
				peak_power.w = pp.w;
				pp=power[ul_PoT+(ul_FftLength>>2)*binsAboveThreshold[n].w];
				peak_power.w = max(peak_power.w,pp.w);

				p = binsAboveThreshold[i].w;
				while( (power[ul_PoT+(ul_FftLength>>2)*p].w >= triplet_thresh.w) && 
					(p <= (int)(floor(midpoint.w))) ) {    /* Check if there is a pulse "off" in between init and midpoint */
					p++;
				}

				q = (int)(floor(midpoint.w))+1;
				while( (power[ul_PoT+(ul_FftLength>>2)*q].w >= triplet_thresh.w) && (q <= binsAboveThreshold[n].w)) {    /* Check if there is a pulse "off" in between midpoint and end */
					q++;
				}

				if( p >= (int)(floor(midpoint.w)) || q >= binsAboveThreshold[n].w) {
					/* if this pulse doesn't have an "off" between all the three spikes, it's dropped */
				} else if( (midpoint.w - floor(midpoint.w)) > 0.1f ) {    /* if it's spread among two bins */
					if( power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.w].w >= triplet_thresh.w ) {
						if (already_reported_flag >= 2)	{
							 result_flag[0].z=1; // Reporting Error, more than 2 results per PoT, redo the calculations on CPU
							return;
						}
						peak_power.w = max(peak_power.w,power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.w].w);
						float4 tmp;
						tmp.x=peak_power.w/mean_power.w;
						tmp.y=mean_power.w;tmp.z=period.w;tmp.w=midpoint.w;
						triplet_results[ul_PoT+ (2*TOffset+already_reported_flag)*(ul_FftLength) + 3/*1 for .y and so on*/]=tmp; 
						result_flag[0].x= 1; // Mark for download <- VERY RARE SITUATION
						already_reported_flag++;
						//ReportTripletEvent( peak_power/mean_power, mean_power, period, midpoint,time_bin, freq_bin, len_power, power, 1 );
					}

					if( power[ul_PoT+(ul_FftLength>>2)*(int)(midpoint.w+1.f)].w >= triplet_thresh.w ) {
						if (already_reported_flag >= 2)	{
							result_flag[0].z=1; // Reporting Error, more than one result per PoT, redo the calculations on CPU
							return;
						}
						peak_power.w = max(peak_power.w,power[ul_PoT+(ul_FftLength>>2)*(int)(midpoint.w+1.f)].w);
						float4 tmp;
						tmp.x=peak_power.w/mean_power.w;
						tmp.y=mean_power.w;tmp.z=period.w;tmp.w=midpoint.w;
						triplet_results[ul_PoT+ (2*TOffset+already_reported_flag)*(ul_FftLength) + 3/*1 for .y and so on*/]=tmp; 
						result_flag[0].x= 1; // Mark for download <- VERY RARE SITUATION
						already_reported_flag++;
						//ReportTripletEvent(peak_power/mean_power, mean_power, period, midpoint,time_bin, freq_bin, len_power, power, 1 );
					}

				} else {            /* otherwise just check the single midpoint bin */
					if( power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.w].w >= triplet_thresh.w ) {
						if (already_reported_flag >= 2)	{
							result_flag[0].z=1; // Reporting Error, more than one result per PoT, redo the calculations on CPU
							return;
						}

						peak_power.w = max(peak_power.w,power[ul_PoT+(ul_FftLength>>2)*(int)midpoint.w].w);

						float4 tmp;
						tmp.x=peak_power.w/mean_power.w;
						tmp.y=mean_power.w;tmp.z=period.w;tmp.w=midpoint.w;
						triplet_results[ul_PoT+ (2*TOffset+already_reported_flag)*(ul_FftLength) + 3/*1 for .y and so on*/]=tmp; 
						result_flag[0].x= 1; // Mark for download <- VERY RARE SITUATION
						already_reported_flag++;
						//ReportTripletEvent( peak_power/mean_power, mean_power, period, midpoint,time_bin, freq_bin, len_power, power, 1 );
					}
				}
			}
		}
	}
#endif
}


__kernel __attribute__((vec_type_hint(float4)))
void PC_find_triplets_kernel_cl(int ul_FftLength, int len_power, float triplet_thresh_base, int AdvanceBy, int PoTLen,
									  __global float4* PoT,__global uint4* result_flag) {
//R: difference from original kernel: this one just doing fast precheck and relies on CPU for real triplet analysis
//R: this kernel does 4 PoT chunks at once.
//R: cause workitems can write flag in arbitrary order it's not possible to set error code actually (original CUDA code
// missed this fact and tries to return error code from kernel. That is, different variable should be used for setting 
// state flags. In this kernel next convention used:
//If whole flag is zero - there is no errors and no result loading on CPU required, kernel launch was silent.
//If result_flag.x not zero - but all other parts are zero - kernel launch was OK, some triplets found, CPU logging required.
// result_flag.y!=0 - there were too many bins above threshold, CPU reprocessing required
// result_flag.z!=0 -there were too many (>2) triplets found in single PoT chunk, CPU reprocessing required
//R: new in precheck kernel (only this flag can be set in precheck kernel):
// result_flag.w!=0 - >2 bins above threshold - CPU analysys required
//	if(result_flag[0].w==1) return;//R: prev workitem already encountered such situation kernel should abort ASAP
	int ul_PoT = get_global_id(0);//R: 4 PoTs at once!
	int y = get_global_id(1);//R: index of offset chunk
	int fft_len4=ul_FftLength>>2;
	int TOffset = y * AdvanceBy;
	if(TOffset + len_power > PoTLen) {		
		TOffset = PoTLen - len_power;
	}
	__global float4* fp_PulsePot= PoT + ul_PoT + TOffset * (fft_len4);
	// Clear the result array
	int4 numBinsAboveThreshold=(int4)0;
	int i;
	float4 mean_power=(float4)0.f,triplet_thresh=(float4)triplet_thresh_base,pp;

	/* Get all the bins that are above the threshold, and find the power array mean value */
	for( i=0;i<len_power;i++ ) {
		mean_power += fp_PulsePot[i*fft_len4];
	}
	mean_power /= (float4)len_power;
	triplet_thresh*=mean_power;

	for( i=0;i<len_power;i++ ) {
		pp= fp_PulsePot[i*fft_len4];
		if(  pp.x>= triplet_thresh.x ) {
			numBinsAboveThreshold.x++;
		}
		if(  pp.y>= triplet_thresh.y ) {
			numBinsAboveThreshold.y++;
		}
		if(  pp.z>= triplet_thresh.z ) {
			numBinsAboveThreshold.z++;
		}
		if(  pp.w>= triplet_thresh.w ) {
			numBinsAboveThreshold.w++;
		}
	}
	if(numBinsAboveThreshold.x>2 || numBinsAboveThreshold.y>2 || numBinsAboveThreshold.z>2 || numBinsAboveThreshold.w>2){
		result_flag[0].w=1;
	}
}


__kernel __attribute__((vec_type_hint(float4)))
void PC_find_triplets_avg_kernel_cl(int ul_FftLength, int len_power, float triplet_thresh_base, int AdvanceBy, int PoTLen,
									  __global float4* PoT,__global uint4* result_flag,__global float4* PulsePoT_average) {
//R: difference from original kernel: this one just doing fast precheck and relies on CPU for real triplet analysis
//R: this kernel does 4 PoT chunks at once.
//R: cause workitems can write flag in arbitrary order it's not possible to set error code actually (original CUDA code
// missed this fact and tries to return error code from kernel. That is, different variable should be used for setting 
// state flags. In this kernel next convention used:
//If whole flag is zero - there is no errors and no result loading on CPU required, kernel launch was silent.
//If result_flag.x not zero - but all other parts are zero - kernel launch was OK, some triplets found, CPU logging required.
// result_flag.y!=0 - there were too many bins above threshold, CPU reprocessing required
// result_flag.z!=0 -there were too many (>2) triplets found in single PoT chunk, CPU reprocessing required
//R: new inj precheck kernel (only this flag can be set in precheck kernel):
// result_flag.w!=0 - >2 bins above threshold - CPU analysys required
//	if(result_flag[0].w==1) return;//R: prev workitem already encountered such situation kernel should abort ASAP
	int ul_PoT = get_global_id(0);//R: 4 PoTs at once!
	int y = get_global_id(1);//R: index of offset chunk
	int fft_len4=ul_FftLength>>2;
	int TOffset = y * AdvanceBy;
	if(TOffset + len_power > PoTLen) {		
		TOffset = PoTLen - len_power;
	}
	__global float4* fp_PulsePot= PoT + ul_PoT + TOffset * (fft_len4);
	// Clear the result array
	int4 numBinsAboveThreshold=(int4)0;
	int i;
	float4 mean_power=(float4)0.f,triplet_thresh=(float4)triplet_thresh_base,pp;

	/* Get all the bins that are above the threshold, and find the power array mean value */
	for( i=0;i<len_power;i++ ) {
		mean_power += fp_PulsePot[i*fft_len4];
	}
	mean_power /= (float4)len_power;
	PulsePoT_average[ul_PoT+y*fft_len4]=mean_power;//R: this avg will be needed later, at pulse finding
	//R: TODO explore if different kernel needed when no pulse finding involved
	triplet_thresh*=mean_power;

	for( i=0;i<len_power;i++ ) {
		pp= fp_PulsePot[i*fft_len4];
		if(  pp.x>= triplet_thresh.x ) {
			numBinsAboveThreshold.x++;
		}
		if(  pp.y>= triplet_thresh.y ) {
			numBinsAboveThreshold.y++;
		}
		if(  pp.z>= triplet_thresh.z ) {
			numBinsAboveThreshold.z++;
		}
		if(  pp.w>= triplet_thresh.w ) {
			numBinsAboveThreshold.w++;
		}
	}
	if(numBinsAboveThreshold.x>2 || numBinsAboveThreshold.y>2 || numBinsAboveThreshold.z>2 || numBinsAboveThreshold.w>2){
		result_flag[0].w=1;
	}
}



/*
0 1 2 3 4 5 6 7 8 9 10 - 11=di
x x x x x - new di
x x x x x x - tmp0=1+new di=6
0+6 1+7 2+8 3+9 (4+10)=stop+(stop+tmp0) - 5=di
 x   x  -new di, tmp0=1+2=3
(0+6)+(3+9) (1+7)+(4+10)
*/ 
float4 sumtop2(__global float4 *tab, __global float4* dest, int di, int fft_len4, int tmp0) {
//R: this function working with 4 different frequency bins at once as all other pulse finding functions do
	float4 sum, tmax;
	int   i;
	tmax = 0.0f;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * fft_len4;

	for (i = 0; i < di; i++) {
		float4 i1 = one[i * (fft_len4)];
		float4 i2 = two[i * (fft_len4)];
		sum = i1 + i2;
		dest[i * (fft_len4)] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float4 sumtop2f(__global float4 *tab, int di, int fft_len4, int tmp0) {
//R: this function working with 4 different frequency bins at once as all other pulse finding functions do
	float4 sum, tmax;
	int   i;
	tmax = 0.0f;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * fft_len4;

	for (i = 0; i < di; i++) {
		float4 i1 = one[i * (fft_len4)];
		float4 i2 = two[i * (fft_len4)];
		sum = i1 + i2;
		//dest[i * (fft_len4)] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}

float4 sumtop2_2(__global float4 *tab, __global float4* dest, int di, int fft_len4, int tmp0,float4* tmax2) {
//R: this function working with 4 different frequency bins at once as all other pulse finding functions do
//R: also it adds/searches 2 levels of fold (stride add) at once
	float4 sum, tmax,sum2;
	int   i;
	tmax = (float4)0.0f;
	*tmax2=(float4)0.0f;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * fft_len4;
	int stop=di-1;
	for (i = 0; i < stop; i++,stop--) {
		float4 i1 = one[i * (fft_len4)];
		float4 i2 = two[i * (fft_len4)];
		float4 j1=	one[stop * (fft_len4)];
		float4 j2=	two[stop * (fft_len4)];
		sum = i1 + i2;
		tmax = max(tmax, sum);
		sum2=j1+j2;
		tmax = max(tmax, sum2);
		sum+=sum2;
		*tmax2=max(*tmax2,sum);
		dest[i * (fft_len4)] = sum;//R: second stride add stored
				
	}
	return tmax;
}
float4 sumtop2_2f(__global float4 *tab,/* __global float4* dest, */int di, int fft_len4, int tmp0,float4* tmax2) {
//R: this function working with 4 different frequency bins at once as all other pulse finding functions do
//R: also it adds/searches 2 levels of fold (stride add) at once and not store result (final)
	float4 sum, tmax,sum2;
	int   i;
	tmax = (float4)0.0f;
	*tmax2=(float4)0.0f;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * fft_len4;
	int stop=di-1;
	for (i = 0; i < stop; i++,stop--) {
		float4 i1 = one[i * (fft_len4)];
		float4 i2 = two[i * (fft_len4)];
		float4 j1=	one[stop * (fft_len4)];
		float4 j2=	two[stop * (fft_len4)];
		sum = i1 + i2;
		tmax = fmax(tmax, sum);
		sum2=j1+j2;
		tmax = fmax(tmax, sum2);
		sum+=sum2;
		*tmax2=fmax(*tmax2,sum);
		//dest[i * (fft_len4)] = sum;//R: second stride add stored
	}
	return tmax;
}
//(i,i+1,i+2,i+3,...,stop-2,stop-1,stop)
//0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17
//di=
float4 sumtop2_3f(__global float4 *tab,/* __global float4* dest, */int di, int fft_len4, int tmp0,float4* tmax2,float4* tmax3) {
//R: this function working with 4 different frequency bins at once as all other pulse finding functions do
//R: also it adds/searches 2 levels of fold (stride add) at once and not store result (final)
	float4 sum,sum1,tmax,sum2,sum3;
	int   i;
	tmax = (float4)0.0f;
	*tmax2=(float4)0.0f;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * fft_len4;
	int stop=di-1;
	for (i = 0; i < stop; i+=2,stop-=2) {
		float4 i1 = one[i * (fft_len4)];
		float4 i2 = two[i * (fft_len4)];
		float4 i1_1 = one[(i+1) * (fft_len4)];
		float4 i2_1 = two[(i+1) * (fft_len4)];

		float4 j1_1=	one[(stop-1) * (fft_len4)];
		float4 j2_1=	two[(stop-1) * (fft_len4)];
		float4 j1=	one[stop * (fft_len4)];
		float4 j2=	two[stop * (fft_len4)];
		sum = i1 + i2;
		tmax = fmax(tmax, sum);
		sum1 = i1_1 + i2_1;
		tmax = fmax(tmax, sum1);
		sum2=j1+j2;
		tmax = fmax(tmax, sum2);
		sum3=j1_1+j2_1;
		tmax = fmax(tmax, sum3);
		sum+=sum2;
		*tmax2=fmax(*tmax2,sum);
		sum1+=sum3;
		//dest[i * (fft_len4)] = sum;//R: second stride add stored
	}
	return tmax;
}
float4 sumtop3(__global float4 *tab, __global float4* dest, int di, int fft_len4, int tmp0, int tmp1) {
	float4 sum, tmax;
	int   i;
	tmax = 0.0f;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * fft_len4;
	__global float4 *three = tab + tmp1 * (fft_len4);

	for (i = 0; i < di; i++) {
		int offset=i * (fft_len4);
		float4 i1 = one[offset];
		float4 i2 = two[offset];
		float4 i3 = three[offset];
		sum = i1 + i2;
		sum += i3;
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float4 sumtop3f(__global float4 *tab, int di, int fft_len4, int tmp0, int tmp1) {
	float4 sum, tmax;
	int   i;
	tmax = 0.0f;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * fft_len4;
	__global float4 *three = tab + tmp1 * (fft_len4);

	for (i = 0; i < di; i++) {
		float4 i1 = one[i * (fft_len4)];
		float4 i2 = two[i * (fft_len4)];
		float4 i3 = three[i * (fft_len4)];
		sum = i1 + i2;
		sum += i3;
		//dest[i * (fft_len4)] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}

float4 sumtop4(__global float4 *tab, __global float4* dest, int di, int fft_len4, int tmp0, int tmp1, int tmp2) {
	float4 sum, tmax;
	int   i;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * (fft_len4);
	__global float4 *three = tab + tmp1 * (fft_len4);
	__global float4 *four = tab + tmp2 * (fft_len4);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		float4 i1 = one[i * fft_len4];
		float4 i2 = two[i * fft_len4];
		float4 i3 = three[i * fft_len4];
		float4 i4 = four[i * fft_len4];
		sum = i1 + i2;
		sum += i3;
		sum += i4;		
		dest[i * fft_len4] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}

float4 sumtop4f(__global float4 *tab, int di, int fft_len4, int tmp0, int tmp1, int tmp2) {
	float4 sum, tmax;
	int   i;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * (fft_len4);
	__global float4 *three = tab + tmp1 * (fft_len4);
	__global float4 *four = tab + tmp2 * (fft_len4);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		float4 i1 = one[i * fft_len4];
		float4 i2 = two[i * fft_len4];
		float4 i3 = three[i * fft_len4];
		float4 i4 = four[i * fft_len4];
		sum = i1 + i2;
		sum += i3;
		sum += i4;		
		//dest[i * fft_len4] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}

float4 sumtop5(__global float4 *tab, __global float4* dest, int di, int fft_len4, int tmp0, int tmp1, int tmp2, int tmp3) {
	float4 sum, tmax;
	int   i;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * (fft_len4);
	__global float4 *three = tab + tmp1 * (fft_len4);
	__global float4 *four = tab + tmp2 * (fft_len4);
	__global float4 *five = tab + tmp3 * (fft_len4);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		float4 i1 = one[i * fft_len4];
		float4 i2 = two[i * fft_len4];
		float4 i3 = three[i * fft_len4];
		float4 i4 = four[i * fft_len4];
		float4 i5 = five[i * fft_len4];
		sum = i1 + i2;
		sum += i3;
		sum += i4;
		sum += i5;
		dest[i * fft_len4] = sum;
		tmax = max(tmax, sum);		
	}	
	return tmax;
}
float4 sumtop5f(__global float4 *tab, int di, int fft_len4, int tmp0, int tmp1, int tmp2, int tmp3) {
	float4 sum, tmax;
	int   i;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * (fft_len4);
	__global float4 *three = tab + tmp1 * (fft_len4);
	__global float4 *four = tab + tmp2 * (fft_len4);
	__global float4 *five = tab + tmp3 * (fft_len4);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		float4 i1 = one[i * fft_len4];
		float4 i2 = two[i * fft_len4];
		float4 i3 = three[i * fft_len4];
		float4 i4 = four[i * fft_len4];
		float4 i5 = five[i * fft_len4];
		sum = i1 + i2;
		sum += i3;
		sum += i4;
		sum += i5;
		//dest[i * fft_len4] = sum;
		tmax = max(tmax, sum);		
	}	
	return tmax;
}
typedef struct { 
	//int NumDataPoints;
	// find_triplets
	//float *power_ft;
	//float4* results_ft;
	//result_flag* result_flags_ft; 	

	// find_pulse
	//float* PulsePot_fp;		// Input data
	//float* PulsePot8_fp;	// Input data moved 8 bytes forward for coleased reads
	//float* tmp_pot_fp;		// Temporary array
	//float* best_pot_fp;		// Copy folded pots with best score
	//float* report_pot_fp;	// Copy folded pots for reporting
	//float4* results_fp;		// Additional data for reporting

	//result_find_pulse_flag* result_flags_fp;

	//const float* t_funct_cache_fp; // cached results of cudaAcc_t_funct								   								   
	float rcfg_dis_thresh;
	int PulseMax;								                                      
} PulseFind_t;
#define FOLDS_COUNT 4
#define FOLDS_START 2
float4 t_funct(int di, int num_adds, int j, int PulseMax, __global float* t_funct_cache) {
	return (float4)t_funct_cache[j * PulseMax * FOLDS_COUNT + (num_adds - FOLDS_START) * PulseMax + di];
}

#if 0 //R: currently avg computed in Triplet pre-check kernel and reused in Pulse find
__kernel void compute_PulsePoT_average_kernel_cl(int PulsePotLen, int AdvanceBy,int fft_len,
									  __global float4* PoT,__global float4* PulsePoT_average){
	int ul_PoT = get_global_id(0);
	int y = get_global_id(1);
	int TOffset1 = y * AdvanceBy;
	const int PoTLen = 1024*1024/fft_len;
	int fft_len4=fft_len>>2;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}
	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	float4 avg=(float4)0.f;
	//  Calculate average power
	for (int i = 0; i < PulsePotLen; i++) {
		avg += fp_PulsePot[i * (fft_len4)];
	}
	avg /= PulsePotLen;
	PulsePoT_average[ul_PoT+y*fft_len4]=avg;
}
#endif
__kernel __attribute__((vec_type_hint(float4)))
void PC_find_pulse_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,int pass,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global float4* tmp_PoT,__global uint4* result_flag,int offset,
								  __global float4* PulsePoT_average) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0)+offset;
	int y = get_global_id(1);
	int z=get_global_id(2);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = (32*y+z) * AdvanceBy;
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg;//,maxp=0;//, snr=0, fthresh=0;
	float4 tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len4];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	int num_adds = pass;
	//for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3-z;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4-z;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5-z;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p-=32) {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr.x  > best_pulse_score ||
					_snr.y  > best_pulse_score ||
					_snr.z  > best_pulse_score ||
					_snr.w  > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			float4 tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				{
					tmp_max = sumtop2(tmp_pot, tmp_pot, di, fft_len4, tmp0);
				}
				if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y ||
					tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr.x  > best_pulse_score ||
						_snr.y  > best_pulse_score ||
						_snr.z  > best_pulse_score ||
						_snr.w  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max.x>cur_thresh.x) ||
						(tmp_max.y>cur_thresh.y) ||
						(tmp_max.z>cur_thresh.z) ||
						(tmp_max.w>cur_thresh.w) ) {
						result_flag[0].x = 1;//return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}
__kernel __attribute__((vec_type_hint(float4)))
void PC_find_pulse_kernel1_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,int pass,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global float4* tmp_PoT,__global uint4* result_flag,int offset,
								  __global float4* PulsePoT_average) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0)+offset;
	int y = get_global_id(1);
	int z=get_global_id(2);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = (get_global_size(2)*y+z) * (PulsePotLen/pass);
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg;//,maxp=0;//, snr=0, fthresh=0;
	float4 tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len4];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	int num_adds = pass;
	//for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3-z;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4-z;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5-z;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p-=get_global_size(2)) {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr.x  > best_pulse_score ||
					_snr.y  > best_pulse_score ||
					_snr.z  > best_pulse_score ||
					_snr.w  > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			float4 tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				{
					tmp_max = sumtop2(tmp_pot, tmp_pot, di, fft_len4, tmp0);
				}
				if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y ||
					tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr.x  > best_pulse_score ||
						_snr.y  > best_pulse_score ||
						_snr.z  > best_pulse_score ||
						_snr.w  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max.x>cur_thresh.x) ||
						(tmp_max.y>cur_thresh.y) ||
						(tmp_max.z>cur_thresh.z) ||
						(tmp_max.w>cur_thresh.w) ) {
						result_flag[0].x = 1;//return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}
__kernel __attribute__((vec_type_hint(float4)))
void PC_find_pulse_partial_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,int pass,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global float4* tmp_PoT,__global uint4* result_flag,int offset,
								  __global float4* PulsePoT_average,int P,int PP) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0)+offset;
	int y = get_global_id(1);
	int z=get_global_id(2);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = (32*y+z) * AdvanceBy;
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg;//,maxp=0;//, snr=0, fthresh=0;
	float4 tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len4];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	int num_adds = pass;
	//for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int lastP;
		switch(num_adds) {
		case 3: lastP = P-z; break;
		case 4: lastP = P-z; break;
		case 5: lastP = P-z; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > PP ; p-=32) {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr.x  > best_pulse_score ||
					_snr.y  > best_pulse_score ||
					_snr.z  > best_pulse_score ||
					_snr.w  > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			float4 tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				{
					tmp_max = sumtop2(tmp_pot, tmp_pot, di, fft_len4, tmp0);
				}
				if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y ||
					tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr.x  > best_pulse_score ||
						_snr.y  > best_pulse_score ||
						_snr.z  > best_pulse_score ||
						_snr.w  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max.x>cur_thresh.x) ||
						(tmp_max.y>cur_thresh.y) ||
						(tmp_max.z>cur_thresh.z) ||
						(tmp_max.w>cur_thresh.w) ) {
						result_flag[0].x = 1;//return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}
__kernel void PC_find_pulse_partial_kernel1_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,int pass,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global float4* tmp_PoT,__global uint4* result_flag,int offset,
								  __global float4* PulsePoT_average,int P,int PP) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0)+offset;
	int y = get_global_id(1);
	int z=get_global_id(2);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = (get_global_size(2)*y+z) * (PulsePotLen/pass);
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg;//,maxp=0;//, snr=0, fthresh=0;
	float4 tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len4];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	int num_adds = pass;
	//for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int lastP;
		switch(num_adds) {
		case 3: lastP = P-z; break;
		case 4: lastP = P-z; break;
		case 5: lastP = P-z; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > PP ; p-=get_global_size(2)) {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr.x  > best_pulse_score ||
					_snr.y  > best_pulse_score ||
					_snr.z  > best_pulse_score ||
					_snr.w  > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			float4 tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				{
					tmp_max = sumtop2(tmp_pot, tmp_pot, di, fft_len4, tmp0);
				}
				if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y ||
					tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr.x  > best_pulse_score ||
						_snr.y  > best_pulse_score ||
						_snr.z  > best_pulse_score ||
						_snr.w  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max.x>cur_thresh.x) ||
						(tmp_max.y>cur_thresh.y) ||
						(tmp_max.z>cur_thresh.z) ||
						(tmp_max.w>cur_thresh.w) ) {
						result_flag[0].x = 1;//return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}

#if 0
float4 sumtop2_team(__global float4 *tab, __global float4* dest, int di, int fft_len4, int tmp0) {
//R: this function working with 4 different frequency bins at once as all other pulse finding functions do
	float4 sum, tmax;
	int   i;
	tmax = (float4)0.0f;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * fft_len4;

	for (i = get_local_id(0); i < di; i+=64) {
		float4 i1 = one[i * (fft_len4)];
		float4 i2 = two[i * (fft_len4)];
		sum = i1 + i2;
		dest[i * (fft_len4)] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float4 sumtop3_team(__global float4 *tab, __global float4* dest, int di, int fft_len4, int tmp0, int tmp1) {
	float4 sum, tmax;
	int   i;
	tmax = (float4)0.0f;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * fft_len4;
	__global float4 *three = tab + tmp1 * (fft_len4);

	for (i = get_local_id(0); i < di; i+=64) {
		float4 i1 = one[i * (fft_len4)];
		float4 i2 = two[i * (fft_len4)];
		float4 i3 = three[i * (fft_len4)];
		sum = i1 + i2;
		sum += i3;
		dest[i * (fft_len4)] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float4 sumtop4_team(__global float4 *tab, __global float4* dest, int di, int fft_len4, int tmp0, int tmp1, int tmp2) {
	float4 sum, tmax;
	int   i;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * (fft_len4);
	__global float4 *three = tab + tmp1 * (fft_len4);
	__global float4 *four = tab + tmp2 * (fft_len4);
	tmax =(float4) 0.0f;

	for (i = get_local_id(0); i < di; i+=64) {
		float4 i1 = one[i * fft_len4];
		float4 i2 = two[i * fft_len4];
		float4 i3 = three[i * fft_len4];
		float4 i4 = four[i * fft_len4];
		sum = i1 + i2;
		sum += i3;
		sum += i4;		
		dest[i * fft_len4] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}

float4 sumtop5_team(__global float4 *tab, __global float4* dest, int di, int fft_len4, int tmp0, int tmp1, int tmp2, int tmp3) {
	float4 sum, tmax;
	int   i;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * (fft_len4);
	__global float4 *three = tab + tmp1 * (fft_len4);
	__global float4 *four = tab + tmp2 * (fft_len4);
	__global float4 *five = tab + tmp3 * (fft_len4);
	tmax = (float4)0.0f;

	for (i = get_local_id(0); i < di; i+=64) {
		float4 i1 = one[i * fft_len4];
		float4 i2 = two[i * fft_len4];
		float4 i3 = three[i * fft_len4];
		float4 i4 = four[i * fft_len4];
		float4 i5 = five[i * fft_len4];
		sum = i1 + i2;
		sum += i3;
		sum += i4;
		sum += i5;
		dest[i * fft_len4] = sum;
		tmax = max(tmax, sum);		
	}	
	return tmax;
}

float4 reduce_max(__local float4* shared_max){
	const int tid = get_local_id(0);		

		barrier(CLK_LOCAL_MEM_FENCE);//write_mem_fence(CLK_LOCAL_MEM_FENCE);
		if (tid < 32) {
			shared_max[tid] = max(shared_max[tid], shared_max[tid + 32]);
		}
		barrier(CLK_LOCAL_MEM_FENCE);//write_mem_fence(CLK_LOCAL_MEM_FENCE);
		if (tid < 16){ 
			shared_max[tid] = max(shared_max[tid], shared_max[tid + 16]);
		}
		barrier(CLK_LOCAL_MEM_FENCE);//write_mem_fence(CLK_LOCAL_MEM_FENCE);
		if (tid < 8){ 
			shared_max[tid] = max(shared_max[tid], shared_max[tid + 8]);
		}
		barrier(CLK_LOCAL_MEM_FENCE);//write_mem_fence(CLK_LOCAL_MEM_FENCE);
		if (tid < 4){ 
			shared_max[tid] = max(shared_max[tid], shared_max[tid + 4]);
		}
		barrier(CLK_LOCAL_MEM_FENCE);//write_mem_fence(CLK_LOCAL_MEM_FENCE);
		if (tid < 2){ 
			shared_max[tid] = max(shared_max[tid], shared_max[tid + 2]);
		}
		barrier(CLK_LOCAL_MEM_FENCE);//write_mem_fence(CLK_LOCAL_MEM_FENCE);
		return max(shared_max[0], shared_max[1]);

}
__kernel __attribute__((vec_type_hint(float4)))
void PC_find_pulse_partial_team_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,int pass,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global float4* tmp_PoT,__global uint4* result_flag,int offset,
								  __global float4* PulsePoT_average,int P,int PP) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_group_id(0);//R: each group works with single PoT4
	int y = get_global_id(1);
	int z=get_global_id(2);
	const uint tid=get_local_id(0);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = (32*y+z) * AdvanceBy;
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;
	__local float4 shared_max[64];//R: depends from how many threads per workgroup possible
	__local float4 shared_avg;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	//float4 avg;//,maxp=0;//, snr=0, fthresh=0;
	float4 tmp_max;
	if(tid==0)	shared_avg=PulsePoT_average[ul_PoT+y*fft_len4];//R:only first workitem in group fetches memory
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	int num_adds = pass;
	//for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int lastP;
		switch(num_adds) {
		case 3: lastP = P-z; break;
		case 4: lastP = P-z; break;
		case 5: lastP = P-z; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		barrier(CLK_LOCAL_MEM_FENCE);//write_mem_fence(CLK_LOCAL_MEM_FENCE);
		for (p = lastP ; p > PP ; p-=32) {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * shared_avg;

			switch(num_adds) {
			case 3:
				shared_max[tid] = sumtop3_team(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				shared_max[tid] = sumtop4_team(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				shared_max[tid] = sumtop5_team(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}
			tmp_max=reduce_max(shared_max);
			if (tid==0 && (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w)) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - shared_avg) * rcfg_dis_thresh + shared_avg;

				float4 _snr = (tmp_max-shared_avg)/(cur_thresh-shared_avg);
				if (_snr.x  > best_pulse_score ||
					_snr.y  > best_pulse_score ||
					_snr.z  > best_pulse_score ||
					_snr.w  > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			float4 tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct(di, num_adds, j,PulseMax,t_funct_cache_fp) * shared_avg;
				{
					shared_max[tid] = sumtop2(tmp_pot, tmp_pot, di, fft_len4, tmp0);
					tmp_max=reduce_max(shared_max);
				}
				if (tid==0 && (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y ||
					tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w)) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - shared_avg) * rcfg_dis_thresh + shared_avg;

					float4 _snr = (tmp_max-shared_avg)/(cur_thresh-shared_avg);
					if (_snr.x  > best_pulse_score ||
						_snr.y  > best_pulse_score ||
						_snr.z  > best_pulse_score ||
						_snr.w  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max.x>cur_thresh.x) ||
						(tmp_max.y>cur_thresh.y) ||
						(tmp_max.z>cur_thresh.z) ||
						(tmp_max.w>cur_thresh.w) ) {
						result_flag[0].x = 1;//return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}
#endif
#if 0
__kernel void PC_find_pulse_kernel2_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,int pass,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global float4* tmp_PoT,__global uint4* result_flag,int offset,
								  __global float4* PulsePoT_average,__global float4* tmp2) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0)+offset;
	int y = get_global_id(1);
	int z=get_global_id(2);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = (32*y+z) * AdvanceBy;
	if(z>=32){
		TOffset2 = (32*y+(z-32)) * AdvanceBy;
	}
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	if(z>=32){//R: other half will use secondary bufer. It';s because of 128MB limit per buffer in current OpenCL ATi implementation
		tmp_pot=tmp2 + ul_PoT + TOffset2 * (fft_len4);
	}
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg;//,maxp=0;//, snr=0, fthresh=0;
	float4 tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len4];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	int num_adds = pass;
	//for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3-z;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4-z;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5-z;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p-=64) {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr.x  > best_pulse_score ||
					_snr.y  > best_pulse_score ||
					_snr.z  > best_pulse_score ||
					_snr.w  > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			float4 tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				{
					tmp_max = sumtop2(tmp_pot, tmp_pot, di, fft_len4, tmp0);
				}
				if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y ||
					tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr.x  > best_pulse_score ||
						_snr.y  > best_pulse_score ||
						_snr.z  > best_pulse_score ||
						_snr.w  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max.x>cur_thresh.x) ||
						(tmp_max.y>cur_thresh.y) ||
						(tmp_max.z>cur_thresh.z) ||
						(tmp_max.w>cur_thresh.w) ) {
						result_flag[0].x = 1;//return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}
__kernel void PC_find_pulse_ndivs2_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global float4* tmp_PoT,__global uint4* result_flag,
								  __global float4* PulsePoT_average) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0);
	int y = get_global_id(1);
	int z = get_global_id(2);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = (32*y+z) * AdvanceBy;
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg=(float4)0.f;
	float4 tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len4];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	//int num_adds = pass;
	for(int num_adds = 3; num_adds <= 5; num_adds++)//R: hope it will be fast enough to do all in one call 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3-z;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4-z;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5-z;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p-=32) {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr.x  > best_pulse_score ||
					_snr.y  > best_pulse_score ||
					_snr.z  > best_pulse_score ||
					_snr.w  > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//int j = 1;
			float4 tmp_max2=(float4)0.f;
			//for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct(di, num_adds, 1,PulseMax,t_funct_cache_fp) * avg;
				{
					tmp_max = sumtop2f(tmp_pot,/* tmp_pot, */di, fft_len4, tmp0);
				}
				if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y ||
					tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr.x  > best_pulse_score ||
						_snr.y  > best_pulse_score ||
						_snr.z  > best_pulse_score ||
						_snr.w  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max.x>cur_thresh.x) ||
						(tmp_max.y>cur_thresh.y) ||
						(tmp_max.z>cur_thresh.z) ||
						(tmp_max.w>cur_thresh.w) ) {
						result_flag[0].x = 1;//return;
					}
				}

				//num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}
__kernel void PC_find_pulse_ndivs3_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global float4* tmp_PoT,__global uint4* result_flag,
								  __global float4* PulsePoT_average) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0);
	int y = get_global_id(1);
	int z = get_global_id(2);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = (32*y+z) * AdvanceBy;
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg=(float4)0.f;
	float4 tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len4];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	//int num_adds = pass;
	for(int num_adds = 3; num_adds <= 5; num_adds++)//R: hope it will be fast enough to do all in one call 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3-z;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4-z;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5-z;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p-=32) {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr.x  > best_pulse_score ||
					_snr.y  > best_pulse_score ||
					_snr.z  > best_pulse_score ||
					_snr.w  > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//int j = 1;
			float4 tmp_max2=(float4)0.f;
			//for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct(di, num_adds, 1,PulseMax,t_funct_cache_fp) * avg;
				float4 dis_thresh2= t_funct(di>>1, num_adds, 2,PulseMax,t_funct_cache_fp) * avg;
				tmp_max= sumtop2_2f(tmp_pot, di, fft_len4, tmp0,&tmp_max2);
				if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y ||
					tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w ||
					tmp_max2.x>dis_thresh2.x || tmp_max2.y>dis_thresh2.y ||
					tmp_max2.z>dis_thresh2.z || tmp_max2.w>dis_thresh2.w 
					) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					tmp_max2/= (num_adds_2<<1);
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;
					float4 cur_thresh2 = (dis_thresh2 / (num_adds_2<<1) - avg) * rcfg_dis_thresh + avg;
					if ((tmp_max.x>cur_thresh.x) ||
						(tmp_max.y>cur_thresh.y) ||
						(tmp_max.z>cur_thresh.z) ||
						(tmp_max.w>cur_thresh.w) ||
						(tmp_max2.x>cur_thresh2.x) ||
						(tmp_max2.y>cur_thresh2.y) ||
						(tmp_max2.z>cur_thresh2.z) ||
						(tmp_max2.w>cur_thresh2.w)) {
						result_flag[0].x = 1;//return;
					}
				}

				//num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}
#endif
__kernel __attribute__((vec_type_hint(float4)))
void PC_find_pulse_f_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global uint4* result_flag,
								  __global float4* PulsePoT_average) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	int z=get_global_id(2);
	int firstP = (PulsePotLen*1)/2;
	int p = (PulsePotLen*2)/3-z;
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0);
	int y = get_global_id(1);

	int TOffset1 = y * AdvanceBy;
	//int TOffset2 = y * AdvanceBy;
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	//__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg;//,maxp=0;//, snr=0, fthresh=0;
	float4 tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len4];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		switch(num_adds) {
		case 4: p = (PulsePotLen*3)/4-z;  firstP = (PulsePotLen*3)/5; 
			break;
		case 5: p = (PulsePotLen*4)/5-z;  firstP = (PulsePotLen*4)/6; 
			break;
		}
		if(p<=firstP) return;//R: return should be safe cause first iteration has biggest number of periods involved
		int num_adds_minus1 = num_adds - 1;
		//int p = lastP;
		/*for (p = lastP ; p > firstP ; p-=)*/ {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3f(fp_PulsePot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4f(fp_PulsePot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5f(fp_PulsePot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr.x  > best_pulse_score ||
					_snr.y  > best_pulse_score ||
					_snr.z  > best_pulse_score ||
					_snr.w  > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					result_flag[0].x = 1;//return;
				}
			}
		} // for (p = lastP
	} // for(num_adds =
}


#if 0
__kernel void PC_find_pulse_ndivs3_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,int pass,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global float4* tmp_PoT,__global uint4* result_flag,int offset,
								  __global float4* PulsePoT_average) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0)+offset;
	int y = get_global_id(1);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = y * AdvanceBy;
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg=0;//,maxp=0;//, snr=0, fthresh=0;
	float4 tmp_max;
#define USE_AVG_CACHE 1
#if !USE_AVG_CACHE
	int i;

	//  Calculate average power
	for (i = 0; i < PulsePotLen; i++) {
		avg += fp_PulsePot[i * (fft_len4)];
	}
	avg /= PulsePotLen;
#else
	avg=PulsePoT_average[ul_PoT+y*fft_len4];
#endif
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	int num_adds = pass;
	//for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p--) {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float4 _snr = (tmp_max-avg)*(float4)sqrt((float)num_adds)/avg;
				float4 _thresh = (cur_thresh-avg)*(float4)sqrt((float)num_adds)/avg;
				if (_snr.x / _thresh.x > best_pulse_score ||
					_snr.y / _thresh.y > best_pulse_score ||
					_snr.z / _thresh.z > best_pulse_score ||
					_snr.w / _thresh.w > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			float4 tmp_max2=(float4)0.f;
			//for (int j = 1; j < ndivs ; j++) 
			{
				//j.x=1;j.y=2;
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct(di, num_adds, 1,PulseMax,t_funct_cache_fp) * avg;
				float4 dis_thresh2= t_funct(di>>1, num_adds, 2,PulseMax,t_funct_cache_fp) * avg;
				tmp_max= sumtop2_2f(tmp_pot, tmp_pot, di, fft_len4, tmp0,&tmp_max2);
				if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y ||
					tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w ||
					tmp_max2.x>dis_thresh2.x || tmp_max2.y>dis_thresh2.y ||
					tmp_max2.z>dis_thresh2.z || tmp_max2.w>dis_thresh2.w 
					) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					tmp_max2/= (num_adds_2<<1);
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;
					float4 cur_thresh2 = (dis_thresh2 / (num_adds_2<<1) - avg) * rcfg_dis_thresh + avg;
					float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
					float4 _snr2 = (tmp_max2-avg)/(cur_thresh2-avg);

					if (_snr.x  > best_pulse_score ||
						_snr.y  > best_pulse_score ||
						_snr.z  > best_pulse_score ||
						_snr.w  > best_pulse_score ||
						_snr2.x  > best_pulse_score ||
						_snr2.y  > best_pulse_score ||
						_snr2.z  > best_pulse_score ||
						_snr2.w  > best_pulse_score ) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max.x>cur_thresh.x) ||
						(tmp_max.y>cur_thresh.y) ||
						(tmp_max.z>cur_thresh.z) ||
						(tmp_max.w>cur_thresh.w) ||
						(tmp_max2.x>cur_thresh2.x) ||
						(tmp_max2.y>cur_thresh2.y) ||
						(tmp_max2.z>cur_thresh2.z) ||
						(tmp_max2.w>cur_thresh2.w)) {
						result_flag[0].x = 1;//return;
					}
				}

				//num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}


__kernel void PC_find_pulse_ndivs3_nb_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,int pass,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global float4* tmp_PoT,__global uint4* result_flag,int offset,
								  __global float4* PulsePoT_average) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0)+offset;
	int y = get_global_id(1);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = y * AdvanceBy;
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg=0;//,maxp=0;//, snr=0, fthresh=0;
	float4 tmp_max;
#define USE_AVG_CACHE 1
#if !USE_AVG_CACHE
	int i;

	//  Calculate average power
	for (i = 0; i < PulsePotLen; i++) {
		avg += fp_PulsePot[i * (fft_len4)];
	}
	avg /= PulsePotLen;
#else
	avg=PulsePoT_average[ul_PoT+y*fft_len4];
#endif
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	int num_adds = pass;
	//for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p--) {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			float4 tmp_max2=(float4)0.f;
			//for (int j = 1; j < ndivs ; j++) 
			{
				//j.x=1;j.y=2;
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct(di, num_adds, 1,PulseMax,t_funct_cache_fp) * avg;
				float4 dis_thresh2= t_funct(di>>1, num_adds, 2,PulseMax,t_funct_cache_fp) * avg;
				tmp_max= sumtop2_2f(tmp_pot, tmp_pot, di, fft_len4, tmp0,&tmp_max2);
				if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y ||
					tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w ||
					tmp_max2.x>dis_thresh2.x || tmp_max2.y>dis_thresh2.y ||
					tmp_max2.z>dis_thresh2.z || tmp_max2.w>dis_thresh2.w 
					) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					tmp_max2/= (num_adds_2<<1);
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;
					float4 cur_thresh2 = (dis_thresh2 / (num_adds_2<<1) - avg) * rcfg_dis_thresh + avg;
					if ((tmp_max.x>cur_thresh.x) ||
						(tmp_max.y>cur_thresh.y) ||
						(tmp_max.z>cur_thresh.z) ||
						(tmp_max.w>cur_thresh.w) ||
						(tmp_max2.x>cur_thresh2.x) ||
						(tmp_max2.y>cur_thresh2.y) ||
						(tmp_max2.z>cur_thresh2.z) ||
						(tmp_max2.w>cur_thresh2.w)) {
						result_flag[0].x = 1;//return;
					}
				}
				//num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}

#endif

#if USE_OPENCL_NV || USE_OPENCL_HD5xxx
float sumtop2_local(__local float *tab, __local float* dest, int di,int tmp0) {
//R: this function working with 4 different frequency bins at once as all other pulse finding functions do
	float sum, tmax;
	int   i;
	tmax = 0.0f;
	__local float *one = tab;
	__local float *two = tab + tmp0*get_local_size(0);

	for (i = 0; i < di; i++) {
		int offset=i*get_local_size(0);
		float i1 = one[offset];
		float i2 = two[offset];
		sum = i1 + i2;
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float sumtop3_local( __global float *tab, __local float* dest, int di, int tmp0, int tmp1,int fftlen) {
	float sum, tmax;
	int   i;
	tmax = 0.0f;
	__global float *one = tab;
	__global float *two = tab + tmp0*fftlen;
	__global float *three = tab + tmp1*fftlen;

	for (i = 0; i < di; i++) {
		int offset=i*fftlen;
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		sum = i1 + i2;
		sum += i3;
		dest[i*get_local_size(0)] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float sumtop4_local(__global float *tab, __local float* dest, int di, int tmp0, int tmp1, int tmp2,int fftlen) {
	float sum, tmax;
	int   i;
	__global float *one = tab;
	__global float *two = tab + tmp0*fftlen;
	__global float *three = tab + tmp1*fftlen;
	__global float *four = tab + tmp2*fftlen;
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		int offset=i*fftlen;
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		float i4 = four[offset];
		sum = i1 + i2;
		sum += i3;
		sum += i4;		
		dest[i*get_local_size(0)] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float sumtop5_local(__global float *tab, __local float* dest, int di, int tmp0, int tmp1, int tmp2, int tmp3,int fftlen) {
	float sum, tmax;
	int   i;
	__global float *one = tab;
	__global float *two = tab + tmp0*fftlen;
	__global float *three = tab + tmp1*fftlen;
	__global float *four = tab + tmp2*fftlen;
	__global float *five = tab + tmp3*fftlen;
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		int offset=i*fftlen;
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		float i4 = four[offset];
		float i5 = five[offset];
		sum = i1 + i2;
		sum += i3;
		sum += i4;
		sum += i5;
		dest[i*get_local_size(0)] = sum;
		tmax = max(tmax, sum);		
	}	
	return tmax;
}
float t_funct_f(int di, int num_adds, int j, int PulseMax, __global float* t_funct_cache) {
	return t_funct_cache[j * PulseMax * FOLDS_COUNT + (num_adds - FOLDS_START) * PulseMax + di];
}
//R: this kernel should be used for small power arrays that fit in local memory comletely
__kernel void PC_find_pulse_local_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float* PoT,__local float* tmp_PoT,__global uint4* result_flag,
								  __global float* PulsePoT_average) 
{
	//if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0);
	int y = get_global_id(1);
	int z=get_local_id(0);
	int TOffset1 = y * AdvanceBy;
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len);
	__local float* tmp_pot		= tmp_PoT +z;
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float avg;//,maxp=0;//, snr=0, fthresh=0;
	float tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	//int num_adds = pass;
	for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p--) {
			float cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct_f(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3_local(fp_PulsePot, tmp_pot, di, tmp0, tmp1,fft_len);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4_local(fp_PulsePot, tmp_pot, di, tmp0, tmp1, tmp2,fft_len);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5_local(fp_PulsePot, tmp_pot, di, tmp0, tmp1, tmp2, tmp3,fft_len);
				break;
			}

			if (tmp_max>dis_thresh) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr  > best_pulse_score ) {
					result_flag[0].x = 1;return;					
				}
				if( (tmp_max>cur_thresh) ) {
					result_flag[0].x = 1;return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			//float tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct_f(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				{
					tmp_max = sumtop2_local(tmp_pot, tmp_pot, di,tmp0);
				}
				if (tmp_max>dis_thresh) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max>cur_thresh) ) {
						result_flag[0].x = 1;return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}
# if 1
float sumtop3_full_local( __local float *tab, __local float* dest, int di, int tmp0, int tmp1) {
	float sum, tmax;
	int   i;
	tmax = 0.0f;
	__local float *one = tab;
	__local float *two = tab + tmp0*get_local_size(0);
	__local float *three = tab + tmp1*get_local_size(0);

	for (i = 0; i < di; i++) {
		int offset=i*get_local_size(0);
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		sum = i1 + i2;
		sum += i3;
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float sumtop4_full_local(__local float *tab, __local float* dest, int di, int tmp0, int tmp1, int tmp2) {
	float sum, tmax;
	int   i;
	__local float *one = tab;
	__local float *two = tab + tmp0*get_local_size(0);
	__local float *three = tab + tmp1*get_local_size(0);
	__local float *four = tab + tmp2*get_local_size(0);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		int offset=i*get_local_size(0);
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		float i4 = four[offset];
		sum = i1 + i2;
		sum += i3;
		sum += i4;		
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float sumtop5_full_local(__local float *tab, __local float* dest, int di, int tmp0, int tmp1, int tmp2, int tmp3) {
	float sum, tmax;
	int   i;
	__local float *one = tab;
	__local float *two = tab + tmp0*get_local_size(0);
	__local float *three = tab + tmp1*get_local_size(0);
	__local float *four = tab + tmp2*get_local_size(0);
	__local float *five = tab + tmp3*get_local_size(0);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		int offset=i*get_local_size(0);
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		float i4 = four[offset];
		float i5 = five[offset];
		sum = i1 + i2;
		sum += i3;
		sum += i4;
		sum += i5;
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}	
	return tmax;
}
float sumtop3_f_full_local( __local float *tab,int di, int tmp0, int tmp1) {
	float sum, tmax;
	int   i;
	tmax = 0.0f;
	__local float *one = tab;
	__local float *two = tab + tmp0*get_local_size(0);
	__local float *three = tab + tmp1*get_local_size(0);

	for (i = 0; i < di; i++) {
		int offset=i*get_local_size(0);
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		sum = i1 + i2;
		sum += i3;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float sumtop4_f_full_local(__local float *tab,int di, int tmp0, int tmp1, int tmp2) {
	float sum, tmax;
	int   i;
	__local float *one = tab;
	__local float *two = tab + tmp0*get_local_size(0);
	__local float *three = tab + tmp1*get_local_size(0);
	__local float *four = tab + tmp2*get_local_size(0);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		int offset=i*get_local_size(0);
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		float i4 = four[offset];
		sum = i1 + i2;
		sum += i3;
		sum += i4;		
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float sumtop5_f_full_local(__local float *tab, int di, int tmp0, int tmp1, int tmp2, int tmp3) {
	float sum, tmax;
	int   i;
	__local float *one = tab;
	__local float *two = tab + tmp0*get_local_size(0);
	__local float *three = tab + tmp1*get_local_size(0);
	__local float *four = tab + tmp2*get_local_size(0);
	__local float *five = tab + tmp3*get_local_size(0);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		int offset=i*get_local_size(0);
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		float i4 = four[offset];
		float i5 = five[offset];
		sum = i1 + i2;
		sum += i3;
		sum += i4;
		sum += i5;
		tmax = max(tmax, sum);		
	}	
	return tmax;
}
__kernel void PC_find_pulse_full_local_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float* PoT,__local float* tmp_PoT,__global uint4* result_flag,
								  __global float* PulsePoT_average) 
{
	//if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0);
	int y = get_global_id(1);
	int z=get_local_id(0);
	int TOffset1 = y * AdvanceBy;
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len);
	__local float* initial_tmp=tmp_PoT+z;
	__local float* tmp_pot		= initial_tmp+PulsePotLen*get_local_size(0);

	for(int i=0;i<PulsePotLen;i++) initial_tmp[i*get_local_size(0)]=fp_PulsePot[i*fft_len];
	//barrier(CLK_LOCAL_MEM_FENCE);//R: barrier not needed ause each thread will read only own local memory fragment

	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;
	
	float avg;//,maxp=0;//, snr=0, fthresh=0;
	float tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	//int num_adds = pass;
	for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p--) {
			float cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct_f(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3_full_local(initial_tmp, tmp_pot, di, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4_full_local(initial_tmp, tmp_pot, di, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5_full_local(initial_tmp, tmp_pot, di, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max>dis_thresh) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr  > best_pulse_score ) {
					result_flag[0].x = 1;return;					
				}
				if( (tmp_max>cur_thresh) ) {
					result_flag[0].x = 1;return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			//float tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct_f(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				{
					tmp_max = sumtop2_local(tmp_pot, tmp_pot, di,tmp0);
				}
				if (tmp_max>dis_thresh) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max>cur_thresh) ) {
						result_flag[0].x = 1;return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}
float sumtop3_fl(__global float *tab, __global float* dest, int di, int tmp0, int tmp1, int fft_len) {
	float sum, tmax;
	int   i;
	tmax = 0.0f;
	__global float *one = tab;
	__global float *two = tab + tmp0 * fft_len;
	__global float *three = tab + tmp1 * (fft_len);

	for (i = 0; i < di; i++) {
		int offset=i * (fft_len);
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		sum = i1 + i2;
		sum += i3;
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float sumtop4_fl(__global float *tab, __global float* dest, int di, int tmp0, int tmp1, int tmp2, int fft_len) {
	float sum, tmax;
	int   i;
	__global float *one = tab;
	__global float *two = tab + tmp0 * (fft_len);
	__global float *three = tab + tmp1 * (fft_len);
	__global float *four = tab + tmp2 * (fft_len);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		int offset=i * (fft_len);
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		float i4 = four[offset];
		sum = i1 + i2;
		sum += i3;
		sum += i4;		
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}

float sumtop5_fl(__global float *tab, __global float* dest, int di, int tmp0, int tmp1, int tmp2, int tmp3, int fft_len) {
	float sum, tmax;
	int   i;
	__global float *one = tab;
	__global float *two = tab + tmp0 * (fft_len);
	__global float *three = tab + tmp1 * (fft_len);
	__global float *four = tab + tmp2 * (fft_len);
	__global float *five = tab + tmp3 * (fft_len);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		int offset=i * (fft_len);
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		float i4 = four[offset];
		float i5 = five[offset];
		sum = i1 + i2;
		sum += i3;
		sum += i4;
		sum += i5;
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}	
	return tmax;
}
float sumtop2_fl(__global float *tab, __global float* dest, int di, int tmp0, int fft_len) {
//R: this function working with 4 different frequency bins at once as all other pulse finding functions do
	float sum, tmax;
	int   i;
	tmax = 0.0f;
	__global float *one = tab;
	__global float *two = tab + tmp0 * fft_len;

	for (i = 0; i < di; i++) {
		int offset=i * (fft_len);
		float i1 = one[offset];
		float i2 = two[offset];
		sum = i1 + i2;
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float sumtop2_semi_local(__global float *tab, __local float* dest, int di, int tmp0, int fft_len) {
//R: this function working with 4 different frequency bins at once as all other pulse finding functions do
	float sum, tmax;
	int   i;
	tmax = 0.0f;
	__global float *one = tab;
	__global float *two = tab + tmp0 * fft_len;

	for (i = 0; i < di; i++) {
		int offset=i * (fft_len);
		float i1 = one[offset];
		float i2 = two[offset];
		sum = i1 + i2;
		dest[i*get_local_size(0)] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
__kernel void PC_find_pulse_semi_local_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float* PoT,__global float* tmp_PoT,__global uint4* result_flag,
								  __global float* PulsePoT_average,__local float* local_tmp) 
{
	//if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0);
	if (ul_PoT < 1) return; // Original find_pulse, omits first element
	int y = get_global_id(1);
	int z=get_local_id(0);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = (y) * (PulsePotLen/3);
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;

	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len);
	__global float* tmp_pot=tmp_PoT + ul_PoT + TOffset2 * (fft_len);		
	__local float* ltmp=local_tmp+z;

	int di;
	
	float avg;//,maxp=0;//, snr=0, fthresh=0;
	float tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p--) {
			int local_only=0;
			float cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct_f(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3_fl(fp_PulsePot, tmp_pot, di, tmp0, tmp1,fft_len);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4_fl(fp_PulsePot, tmp_pot, di, tmp0, tmp1, tmp2,fft_len);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5_fl(fp_PulsePot, tmp_pot, di, tmp0, tmp1, tmp2, tmp3,fft_len);
				break;
			}

			if (tmp_max>dis_thresh) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr  > best_pulse_score ) {
					result_flag[0].x = 1;return;					
				}
				if( (tmp_max>cur_thresh) ) {
					result_flag[0].x = 1;return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			//float tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct_f(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				if(local_only)//R: both arrays in local memory already
				{
					tmp_max = sumtop2_local(ltmp, ltmp, di,tmp0);
				}else if(di<=60)//R:new array will fit inside local memory
				{
					local_only=1;
					tmp_max = sumtop2_semi_local(tmp_pot, ltmp, di,tmp0,fft_len);
				}else
				{
					tmp_max = sumtop2_fl(tmp_pot, tmp_pot, di,tmp0,fft_len);
				}
				if (tmp_max>dis_thresh) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max>cur_thresh) ) {
						result_flag[0].x = 1;return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}
__kernel void PC_find_pulse_partial_semi_local_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,int pass,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float* PoT,__global float* tmp_PoT,__global uint4* result_flag,int offset,
								  __global float* PulsePoT_average,int P,int PP,__local float* local_tmp) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0)+offset;
	if (ul_PoT < 1) return; // Original find_pulse, omits first element
	int y = get_global_id(1);
	int z=get_global_id(2);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = (get_global_size(2)*y+z) * (PulsePotLen/pass);
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;

	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len);
	__global float* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len);
	__local float* ltmp=local_tmp+(get_local_id(0)+get_local_id(2)*get_local_size(0));

	int di;

	float avg;//,maxp=0;//, snr=0, fthresh=0;
	float tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	int num_adds = pass;
	//for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int lastP;
		switch(num_adds) {
		case 3: lastP = P-z; break;
		case 4: lastP = P-z; break;
		case 5: lastP = P-z; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > PP ; p-=get_global_size(2)) {
			int local_only=0;
			float cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct_f(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3_fl(fp_PulsePot, tmp_pot, di, tmp0, tmp1,fft_len);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4_fl(fp_PulsePot, tmp_pot, di, tmp0, tmp1, tmp2,fft_len);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5_fl(fp_PulsePot, tmp_pot, di,tmp0, tmp1, tmp2, tmp3,fft_len);
				break;
			}

			if (tmp_max>dis_thresh) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr  > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max>cur_thresh) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			float tmp_max2=0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct_f(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				if(local_only)//R: both arrays in local memory already
				{
					tmp_max = sumtop2_local(ltmp, ltmp, di,tmp0);
				}else if(di<=60)//R:new array will fit inside local memory
				{
					local_only=1;
					tmp_max = sumtop2_semi_local(tmp_pot, ltmp, di,tmp0,fft_len);
				}else
				{
					tmp_max = sumtop2_fl(tmp_pot, tmp_pot, di,tmp0,fft_len);
				}
				if (tmp_max>dis_thresh) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max>cur_thresh) ) {
						result_flag[0].x = 1;//return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}

__kernel void PC_find_pulse_f_full_local_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float* PoT,__local float* tmp_PoT,__global uint4* result_flag,
								  __global float* PulsePoT_average) 
{
	//if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0);
	int y = get_global_id(1);
	int z=get_local_id(0);
	int TOffset1 = y * AdvanceBy;
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len);
	__local float* initial_tmp=tmp_PoT+z;

	for(int i=0;i<PulsePotLen;i++) initial_tmp[i*get_local_size(0)]=fp_PulsePot[i*fft_len];

	int di;
	
	float avg;//,maxp=0;//, snr=0, fthresh=0;
	float tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p--) {
			float cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct_f(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3_f_full_local(initial_tmp, di, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4_f_full_local(initial_tmp, di, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5_f_full_local(initial_tmp, di, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max>dis_thresh) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr  > best_pulse_score ) {
					result_flag[0].x = 1;return;					
				}
				if( (tmp_max>cur_thresh) ) {
					result_flag[0].x = 1;return;
				}
			}
		} // for (p = lastP
	} // for(num_adds =
}
#endif

#endif

#if SETI7
#define UNSTDMAX(a,b)  (((a) > (b)) ? (a) : (b))
#define UNSTDMIN(a,b)  (((a) < (b)) ? (a) : (b))
unsigned int pow2a(unsigned int num)
{
    unsigned int n = num > 0 ? num - 1 : 0;
    n |= n >> 1;
    n |= n >> 2;
    n |= n >> 4;
    n |= n >> 8;
    n |= n >> 16;
    n++;

    return n >> 1;
}

__kernel void FindAutoCorrelation_reduce0_kernel_cl(__global float* AutoCorrelation, __global float* PartialSumMaxIdx){
	const int tid=get_global_id(0);
	const y=get_global_id(1)*(64*1024);
	float partial_sum=0.0f;
	float partial_max=0.0;
	float idx=0.0;
	__global float* auto_corr=AutoCorrelation+y;
	for(int i=tid*256;i<(tid+1)*256;i++){
		float tmp=auto_corr[i];
		tmp*=tmp;
		partial_sum+=tmp;
		if(i!=0 && tmp>partial_max){
			partial_max=tmp;
			idx=(float)i;
		}
	}
	int did=3*(256*get_global_id(1)+tid);
	PartialSumMaxIdx[did]=partial_sum;
	PartialSumMaxIdx[did+1]=partial_max;
	PartialSumMaxIdx[did+2]=idx;
}
__kernel void FindAutoCorrelation_kernel_cl(__global float* AutoCorrelation, __global float2* result,int blksize){
//R: autocorr FFT length hadwired to 128k, hence block size is  and half size is 64k
//R: very direct and unoptimized implementation just to get GPU work for SETI7 too

  const int tid=get_global_id(0);
  int i, j, k, m;
  __global float* auto_corr=AutoCorrelation+tid*64*1024;
  float temp;
  float total, MeanPower, partial;
  i = j = k = m = 0;
  total = 0.0f;
  for(int b = 0; b < (64*1024)/blksize; b++) {
      partial = 0.0f;
	  for(i = 0; i < blksize; i++) {
		auto_corr[b*blksize+i]*=auto_corr[b*blksize+i];
		partial += auto_corr[b*blksize+i];
	  }
	  total += partial;
  }
  MeanPower = total / (128*1024);
    temp = 0.0;
    // Walk the array, looking for the first/next highest power.
    // Start j at 1, in order to skip the DC (ie 0) bin.
    // NOTE: this is a simple scan for high powers.  Nice and fast
    // for a very low i.  If i is substantial, this code should be
    // replaced with an index (q)sort.  

  // Here we extract the autocorrs_to_report highest power events,
  // outputing them as we go.
  // Index usage:
  // i : walk power spectrum us_NumToReport times
  // j : walks power spectrum for each i
  // k : marks current high power candidate while j walks on
  // m : marks the current tail of the high power hit "list"
	
	for (j = 1; j < (64*1024); j++) {

      if (auto_corr[j] > temp) {
        if (auto_corr[j] < auto_corr[m] || m == 0) {
          temp = auto_corr[j];
          k = j;
        }
      }
    } // temp now = first/next highest power and k = it's bin number

    m = k; 		// save the "lowest" highest.
	float power=temp/MeanPower;
	result[tid].x=power;//peak value
	result[tid].y=m;//bin
}
__kernel void RepackOutput_kernel_cl(__global float2* src,__global float2* dst) 
{
	//R: taking Jason's CUDA kernel as base and extend to 2D matrix to process whole icfft at once 
	int y=get_global_id(1);
	int x=get_global_id(0);
	int didx = x+y*get_global_size(0);  //packing into float2s
	int sidx = x*2+y*get_global_size(0)*16;//R: stride of 512k float2 elements and we use only half of each array
//	float scale = 2.0f;	

	float a = src[sidx].x;//*src[sidx].x;
	float b = src[sidx+1].x;//*src[sidx+2].x;

//	a = __fmul_rn(a,scale);
//	b = __fmul_rn(b,scale);

	dst[didx] = (float2)(a,b);
}

#endif